"""
This module implements routines to build the `evt` tier.
"""

from __future__ import annotations

import importlib
import itertools
import logging
import re
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any

import awkward as ak
import numpy as np
from lgdo import Array, ArrayOfEqualSizedArrays, Table, VectorOfVectors, lh5

from ..utils import load_dict
from . import aggregators, utils
from .build_tcm import _concat_tables

log = logging.getLogger(__name__)


def build_evt(
    datainfo: utils.DataInfo | Mapping[str, Sequence[str, ...]],
    config: str | Mapping[str, ...],
    wo_mode: str = "write_safe",
    buffer_len=10**4,
) -> None | Table:
    r"""Transform data from hit-structured tiers to event-structured data.

    Parameters
    ----------
    datainfo
        input and output LH5 datainfo with HDF5 groups where tables are found,
        (see :obj:`.utils.DataInfo`). Example: ::

            # syntax: {"tier-name": ("file-name", "hdf5-group"[, "table-format"])}
            {
              "tcm": ("data-tier_tcm.lh5", "hardware_tcm_1"),
              "dsp": ("data-tier_dsp.lh5", "dsp", "ch{}"),
              "hit": ("data-tier_hit.lh5", "hit", "ch{}"),
              "evt": ("data-tier_evt.lh5", "evt")
            }

    config
        name of configuration file or dictionary defining event fields. Channel
        lists can be defined by importing a metadata module.

        - ``channels`` specifies the channels used to for this field (either a
          string or a list of strings).
        - ``channel_mapping`` specifies a dictionary that maps the channel to a name
        - ``operations`` defines the event fields (``name=key``). If the key
          contains slahes it will be interpreted as the path to the output
          field inside nested sub-tables.
        - ``outputs`` defines the fields that are actually included in the
          output table.

        Inside the ``operations`` block:

        - ``aggregation_mode`` defines how the channels should be combined (see
          :func:`evaluate_expression`).
        - ``expression`` defines the expression or function call to apply
          (see :func:`evaluate_expression`),
        - ``query`` defines an expression to mask the aggregation.
        - ``parameters`` defines any other parameter used in expression.
        - ``dtype`` defines the NumPy data type of the resulting data.
        - ``initial`` defines the initial/default value. Useful with some types
          of aggregators.

        For example:

        .. code-block:: json

            {
              "channels": {
                "geds_on": ["ch1084803", "ch1084804", "ch1121600"],
                "spms_on": ["ch1057600", "ch1059201", "ch1062405"],
                "muon": "ch1027202",
              },
              "channelmap" : {
                "ch1084803": "Gethin",
                "ch1084804": "Gertrude",
                "ch1121600": "Geoffrey",
                "ch1057600": "Simon",
                "ch1059201": "Sinbad",
                "ch1062405": "Silvia",
                "ch1027202": "Mulan"
                },
              "outputs": ["energy_id", "multiplicity"],
              "operations": {
                "energy_id":{
                  "channels": "geds_on",
                  "aggregation_mode": "gather",
                  "query": "hit.cuspEmax_ctc_cal > 25",
                  "expression": "tcm.table_key",
                  "sort": "ascend_by:dsp.tp_0_est"
                },
                "energy":{
                  "aggregation_mode": "keep_at_ch:evt.energy_id",
                  "expression": "hit.cuspEmax_ctc_cal > 25"
                }
                "is_muon_rejected":{
                  "channels": "muon",
                  "aggregation_mode": "any",
                  "expression": "dsp.wf_max > a",
                  "parameters": {"a": 15100},
                  "initial": false
                },
                "multiplicity":{
                  "channels":  ["geds_on", "geds_no_psd", "geds_ac"],
                  "aggregation_mode": "sum",
                  "expression": "hit.cuspEmax_ctc_cal > a",
                  "parameters": {"a": 25},
                  "initial": 0
                },
                "t0":{
                  "aggregation_mode": "keep_at_ch:evt.energy_id",
                  "expression": "dsp.tp_0_est",
                  "initial": "np.nan"
                },
                "lar_energy":{
                  "channels": "spms_on",
                  "aggregation_mode": "function",
                  "expression": "pygama.evt.modules.spms.gather_pulse_data(<...>, observable='hit.energy_in_pe')"
                },
              }
            }

    wo_mode
        writing mode, see :func:`lgdo.lh5.core.write`.
    """
    if not isinstance(config, dict):
        config = load_dict(config)

    if "channels" not in config.keys():
        raise ValueError("channel field needs to be specified in the config")
    if "operations" not in config.keys():
        raise ValueError("operations field needs to be specified in the config")

    # convert into a nice named tuple
    if not isinstance(datainfo, utils.DataInfo):
        datainfo = utils.make_files_config(datainfo)

    # check chname_fmt validity
    chname_fmt = datainfo.hit.table_fmt
    pattern_check = re.findall(r"{([^}]*?)}", chname_fmt)
    if len(pattern_check) != 1:
        raise ValueError("chname_fmt must have exactly one placeholder {}")
    elif "{" in pattern_check[0] or "}" in pattern_check[0]:
        raise ValueError(f"{chname_fmt=} has an invalid placeholder.")
    hit_keys = [
        key
        for key in lh5.ls(datainfo.hit.file)
        if utils.get_tcm_id_by_pattern(chname_fmt, key) is not None
    ]
    if (
        utils.get_table_name_by_pattern(
            chname_fmt,
            utils.get_tcm_id_by_pattern(chname_fmt, hit_keys[0]),
        )
        != hit_keys[0]
    ):
        raise ValueError(f"chname_fmt {chname_fmt} does not match keys in data!")

    # create channel list according to config
    # This can be either read from the meta data
    # or a list of channel names
    log.debug("creating channel dictionary")

    channels = {}

    for key, v in config["channels"].items():
        if isinstance(v, dict):
            # it is a meta module. module_name must exist
            if "module" not in v.keys():
                raise ValueError(
                    "Need module_name to load channel via a meta data module"
                )

            attr = {}
            # the time_key argument is mandatory
            if "time_key" not in v.keys():
                raise RuntimeError("the 'time_key' configuration field is mandatory")

            # if "None" do None
            elif "None" == v["time_key"]:
                attr["time_key"] = None

            # load module
            p, m = v["module"].rsplit(".", 1)
            met = getattr(importlib.import_module(p, package=__package__), m)
            channels[key] = met(v | attr)

        elif isinstance(v, str):
            channels[key] = [v]

        elif isinstance(v, list):
            channels[key] = [e for e in v]

    if "channel_mapping" in config.keys():
        channel_mapping = config["channel_mapping"]
        if isinstance(channel_mapping, str):
            channel_mapping = load_dict(channel_mapping)
    else:
        channel_mapping = None

    evt_tbl = build_evt_cols(
        datainfo, config, channels, wo_mode, buffer_len, channel_mapping
    )
    if datainfo.evt.file is None:
        return evt_tbl


def build_evt_cols(
    datainfo: utils.DataInfo | Mapping[str, Sequence[str, ...]],
    config: dict,
    channels: list,
    wo_mode: str = "write_safe",
    buffer_len=10**4,
    channel_mapping: dict = None,
) -> None | Table:
    """
    Iterates through the TCM file and builds the event table according to the
    configuration file. The event table is written to the output file if an
    evt output file is specified otherwise it is returned.

    Parameters
    ----------
    datainfo
        input and output LH5 datainfo with HDF5 groups where tables are found,
        (see :obj:`.utils.DataInfo`). Example: ::

            # syntax: {"tier-name": ("file-name", "hdf5-group"[, "table-format"])}
            {
              "tcm": ("data-tier_tcm.lh5", "hardware_tcm_1"),
              "dsp": ("data-tier_dsp.lh5", "dsp", "ch{}"),
              "hit": ("data-tier_hit.lh5", "hit", "ch{}"),
              "evt": ("data-tier_evt.lh5", "evt")
            }

    config
        dict as defined in the :func:`build_evt` function.
    channels
        list of channels to be used in the event table.
    wo_mode
        writing mode, see :func:`lgdo.lh5.core.write`.
    buffer_len
        number of rows to be processed at once.
    channel_mapping
        dictionary that maps the channel to a name. This can be used in functions
        to get the channel name instead of the channel number.

    Returns
    -------
    None
        if an evt output file is specified otherwise the event table is returned.
    """

    # convert into a nice named tuple
    if not isinstance(datainfo, utils.DataInfo):
        datainfo = utils.make_files_config(datainfo)

    evt_tables = []
    if datainfo.evt.file is not None and wo_mode == "of":
        if Path(datainfo.evt.file).exists():
            Path(datainfo.evt.file).unlink()

    for tcm_lh5 in lh5.LH5Iterator(
        datainfo.tcm.file,
        datainfo.tcm.group,
        buffer_len=buffer_len,
        field_mask=["table_key", "row_in_table"],
    ):
        # load tcm data from disk
        tcm = utils.TCMData(
            table_key=tcm_lh5.table_key.view_as("ak"),
            row_in_table=tcm_lh5.row_in_table.view_as("ak"),
        )

        # get number of events in file (ask the TCM)
        table = Table(size=len(tcm_lh5))

        # now loop over operations (columns in evt table)
        for field, v in config["operations"].items():
            log.debug(f"processing field: '{field}'")

            # if mode not defined in operation, it can only be an operation on the
            # evt level
            if "aggregation_mode" not in v.keys():
                # compute and eventually get rid of evt. suffix
                obj = table.eval(
                    v["expression"].replace("evt.", ""), v.get("parameters", {})
                )

                # add attributes if present
                if "lgdo_attrs" in v.keys():
                    obj.attrs |= v["lgdo_attrs"]

            # else we build the event entry
            else:
                if "channels" not in v.keys():
                    channels_e = []
                elif isinstance(v["channels"], str):
                    channels_e = channels[v["channels"]]
                elif isinstance(v["channels"], list):
                    channels_e = list(
                        itertools.chain.from_iterable(
                            [channels[e] for e in v["channels"]]
                        )
                    )
                channels_skip = []
                if "exclude_channels" in v.keys():
                    if isinstance(v["exclude_channels"], str):
                        channels_skip = channels[v["exclude_channels"]]
                    elif isinstance(v["exclude_channels"], list):
                        channels_skip = list(
                            itertools.chain.from_iterable(
                                [channels[e] for e in v["exclude_channels"]]
                            )
                        )

                defaultv = v.get("initial", np.nan)
                if isinstance(defaultv, str) and (
                    defaultv in ["np.nan", "np.inf", "-np.inf"]
                ):
                    defaultv = eval(defaultv)

                obj = evaluate_expression(
                    datainfo,
                    tcm,
                    channels=channels_e,
                    channels_skip=channels_skip,
                    mode=v["aggregation_mode"],
                    expr=v["expression"],
                    n_rows=len(tcm_lh5),
                    table=table,
                    parameters=v.get("parameters", None),
                    query=v.get("query", None),
                    default_value=defaultv,
                    sorter=v.get("sort", None),
                    channel_mapping=channel_mapping,
                )

                # add attribute if present
                if "lgdo_attrs" in v.keys():
                    obj.attrs |= v["lgdo_attrs"]

            # cast to type, if required
            # hijack the poor LGDO
            if "dtype" in v:
                type_ = v["dtype"]

                if isinstance(obj, Array):
                    obj.nda = obj.nda.astype(type_)
                if isinstance(obj, VectorOfVectors):
                    fldata_ptr = obj.flattened_data
                    while isinstance(fldata_ptr, VectorOfVectors):
                        fldata_ptr = fldata_ptr.flattened_data

                    fldata_ptr.nda = fldata_ptr.nda.astype(type_)

            log.debug(f"new column {field!s} = {obj!r}")
            table.add_field(field, obj)

        # might need to re-organize fields in subtables, create a new object for that
        nested_tbl = Table(size=len(tcm_lh5))
        output_fields = config.get("outputs", table.keys())

        for field, obj in table.items():
            # also only add fields requested by the user
            if field not in output_fields:
                continue

            # if names contain slahes, put in sub-tables
            lvl_ptr = nested_tbl
            subfields = field.strip("/").split("___")
            for level in subfields:
                # if we are at the end, just add the field
                if level == subfields[-1]:
                    lvl_ptr.add_field(level, obj)
                    break

                if not level:
                    msg = f"invalid field name '{field}'"
                    raise RuntimeError(msg)

                # otherwise, increase nesting
                if level not in lvl_ptr:
                    lvl_ptr.add_field(level, Table(size=len(tcm_lh5)))
                lvl_ptr = lvl_ptr[level]

        # write output fields into outfile
        if output_fields:
            if datainfo.evt.file is None:
                evt_tables.append(nested_tbl)
            else:
                lh5.write(
                    obj=nested_tbl,
                    name=datainfo.evt.group,
                    lh5_file=datainfo.evt.file,
                    wo_mode="o" if wo_mode == "u" else "a",
                )
        else:
            # warning will be given on each iteration, maybe not ideal?
            log.warning("no output fields specified, no file will be written.")
            evt_tables.append(nested_tbl)
        if datainfo.evt.file is None:
            return _concat_tables(evt_tables)


def evaluate_expression(
    datainfo: utils.DataInfo | Mapping[str, Sequence[str, ...]],
    tcm: utils.TCMData,
    channels: Sequence[str],
    channels_skip: Sequence[list],
    mode: str,
    expr: str,
    n_rows: int,
    table: Table = None,
    parameters: Mapping[str, Any] = None,
    query: str = None,
    default_value: bool | int | float = np.nan,
    sorter: str = None,
    channel_mapping: dict = None,
) -> Array | ArrayOfEqualSizedArrays | VectorOfVectors:
    """Evaluates the expression defined by the user across all channels
    according to the mode.

    Parameters
    ----------
    datainfo
        input and output LH5 files with HDF5 groups where tables are found.
        (see :obj:`.utils.DataInfo`)
    tcm
        tcm data structure (see :obj:`.utils.TCMData`)
    channels
       list of channel names across which expression gets evaluated
    channels_skip
       list of channels which get set to default value during evaluation. In
       function mode they are removed entirely
    mode
       The mode determines how the event entry is calculated across channels.
       Options are:

       - ``first_at:sorter``: aggregates across channels by returning the
         expression of the channel with smallest value of sorter.
       - ``last_at``: aggregates across channels by returning the expression of
         the channel with largest value of sorter.
       - ``sum``: aggregates by summation.
       - ``any``: aggregates by logical or.
       - ``all``: aggregates by logical and.
       - ``keep_at_ch:ch_field``: aggregates according to passed ch_field.
       - ``keep_at_idx:tcm_idx_field``: aggregates according to passed tcm
         index field.
       - ``gather``: channels are not combined, but result saved as
         :class:`.VectorOfVectors`.
       - ``function``: the function call specified in `expr` is evaluated, and
         the resulting column is inserted into the output table.

    query
       a query that can mask the aggregation.
    expr
       the expression. That can be any mathematical equation/comparison. If
       `mode` is ``function``, the expression needs to be a special processing
       function defined in :mod:`.modules`. In the expression, parameters from
       either `evt` or lower tiers (from operations performed before this one!
       Dictionary operations order matters), or from the ``parameters`` field
       can be used. Fields can be prefixed with the tier id (e.g.
       ``evt.energy`` or `hit.quality_flag``).
    n_rows
       number of rows to be processed.
    table
       table of `evt` tier data.
    parameters
       dictionary of parameters defined in the ``parameters`` field in the
       configuration dictionary.
    default_value
       default value of evaluation.
    sorter
       can be used to sort vector outputs according to sorter expression (see
       :func:`.evaluate_to_vector`).
    channel_mapping
       dictionary that maps the channel to a name. This can be used in functions
       to get the channel name instead of the channel number.

    Note
    ----
    The specification of custom functions that can be used as expression is
    documented in :mod:`.modules`.
    """
    if not isinstance(datainfo, utils.DataInfo):
        datainfo = utils.make_files_config(datainfo)

    # build dictionary of parameter names and their values
    # a parameter can be a column in the existing table...
    pars_dict = {}

    if table is not None:
        pars_dict = {
            k: v for k, v in table.items() if isinstance(v, (Array, VectorOfVectors))
        }

    # ...or defined through the configuration
    if parameters:
        pars_dict = pars_dict | parameters

    if mode == "function":
        # syntax:
        #
        #     pygama.evt.modules.spms.my_func([...], arg1=val, arg2=val)

        # get arguments list passed to the function (outermost parentheses)
        result = re.search(r"\((.*)\)$", expr.strip(), re.DOTALL)
        if result is None:
            msg = f"could not parse the function arguments in '{expr}'"
            raise RuntimeError(msg)

        args_str = result.group(1)

        # handle tier scoping: evt.<>
        args_str = args_str.replace("evt.", "")

        good_chns = [x for x in channels if x not in channels_skip]

        # replace stuff before first comma with list of mandatory args
        full_args_str = "datainfo, tcm, table_names, channel_mapping," + ",".join(
            args_str.split(",")[1:]
        )

        # get module and function names
        func_call = expr.strip().split("(")[0]
        subpackage, _ = func_call.rsplit(".", 1)
        package = subpackage.split(".")[0]

        # import function into current namespace
        log.debug(f"importing module {subpackage}")
        importlib.import_module(subpackage, package=__package__)

        # declare imported package as globals (see eval() call later)
        globs = {
            package: importlib.import_module(package),
        }

        # lookup dictionary for variables used in function arguments (see eval() call later)
        locs = {
            "datainfo": datainfo,
            "tcm": tcm,
            "table_names": good_chns,
            "channel_mapping": channel_mapping,
        } | pars_dict

        # evil eval() to avoid annoying args casting logic
        call_str = f"{func_call}({full_args_str})"
        log.debug(f"evaluating {call_str}")
        log.debug(f"...globals={globs}")
        log.debug(f"...locals={locs}")

        return eval(call_str, globs, locs)

    else:
        # find parameters in evt file or in parameters
        field_list = re.findall(
            rf"({'|'.join(datainfo._asdict().keys())}).([a-zA-Z_$][\w$]*)", expr
        )

        # check if query is either on channel basis or evt basis (and not a mix)
        query_mask = query
        if query is not None:
            hit_tiers = [k for k in datainfo._asdict() if k != "evt"]
            if "evt." in query and (any([t in query for t in hit_tiers])):
                raise ValueError(
                    f"Query can't be a mix of {datainfo.evt.group} tier and lower tiers."
                )

            # if it is an evt query we can evaluate it directly here
            if table and "evt." in query:
                query_mask = eval(query.replace("evt.", ""), table)

        # switch through modes
        if table and (
            mode.startswith("keep_at_ch:") or mode.startswith("keep_at_idx:")
        ):
            if mode.startswith("keep_at_ch:"):
                ch_comp = table[mode[11:].replace("evt.", "")]
            else:
                ch_comp = table[mode[12:].replace("evt.", "")]
                if isinstance(ch_comp, Array):
                    ch_comp = Array(
                        ak.flatten(tcm.table_key)[ch_comp.view_as("np")].to_numpy()
                    )
                elif isinstance(ch_comp, VectorOfVectors):
                    ch_comp = ch_comp.view_as("ak")
                    ch_comp = VectorOfVectors(
                        ak.unflatten(
                            ak.flatten(tcm.table_key)[ak.flatten(ch_comp)].to_numpy(),
                            ak.count(ch_comp, axis=-1),
                        )
                    )
                else:
                    raise NotImplementedError(
                        type(ch_comp)
                        + " not supported (only Array and VectorOfVectors are supported)"
                    )

            if isinstance(ch_comp, Array):
                return aggregators.evaluate_at_channel(
                    datainfo=datainfo,
                    tcm=tcm,
                    channels=channels,
                    channels_skip=channels_skip,
                    expr=expr,
                    field_list=field_list,
                    ch_comp=ch_comp,
                    pars_dict=pars_dict,
                    default_value=default_value,
                    channel_mapping=channel_mapping,
                )

            if isinstance(ch_comp, VectorOfVectors):
                return aggregators.evaluate_at_channel_vov(
                    datainfo=datainfo,
                    tcm=tcm,
                    expr=expr,
                    field_list=field_list,
                    ch_comp=ch_comp,
                    channels=channels,
                    channels_skip=channels_skip,
                    pars_dict=pars_dict,
                    default_value=default_value,
                    channel_mapping=channel_mapping,
                )

            raise NotImplementedError(
                "{type(ch_comp).__name__} not supported "
                "(only Array and VectorOfVectors are supported)"
            )

        if "first_at:" in mode or "last_at:" in mode:
            sorter = tuple(
                re.findall(
                    rf"({'|'.join(datainfo._asdict().keys())}).([a-zA-Z_$][\w$]*)",
                    mode.split("first_at:")[-1],
                )[0]
            )
            return aggregators.evaluate_to_first_or_last(
                datainfo=datainfo,
                tcm=tcm,
                channels=channels,
                channels_skip=channels_skip,
                expr=expr,
                field_list=field_list,
                query=query_mask,
                n_rows=n_rows,
                sorter=sorter,
                pars_dict=pars_dict,
                default_value=default_value,
                is_first=True if "first_at:" in mode else False,
                channel_mapping=channel_mapping,
            )

        if mode in ["sum", "any", "all"]:
            return aggregators.evaluate_to_scalar(
                datainfo=datainfo,
                tcm=tcm,
                mode=mode,
                channels=channels,
                channels_skip=channels_skip,
                expr=expr,
                field_list=field_list,
                query=query_mask,
                n_rows=n_rows,
                pars_dict=pars_dict,
                default_value=default_value,
                channel_mapping=channel_mapping,
            )
        if mode == "gather":
            return aggregators.evaluate_to_vector(
                datainfo=datainfo,
                tcm=tcm,
                channels=channels,
                channels_skip=channels_skip,
                expr=expr,
                field_list=field_list,
                query=query_mask,
                n_rows=n_rows,
                pars_dict=pars_dict,
                default_value=default_value,
                sorter=sorter,
                channel_mapping=channel_mapping,
            )

        raise ValueError(f"'{mode}' is not a valid mode")



from math import erf, erfc

import numpy as np

import pygama.math.functions.error_function as pgfef


def test_erf():
    x = np.arange(-10, 12)

    y = pgfef.nb_erf(x)
    y_math = []
    for i in x:
        y_math.append(erf(i))
    assert np.array_equal(y, y_math)


def test_erfc():
    x = np.arange(-10, 12)

    y = pgfef.nb_erfc(x)
    y_math = []
    for i in x:
        y_math.append(erfc(i))
    assert np.array_equal(y, y_math)



import numpy as np
from pytest import approx

from pygama.math.functions.hpge_peak import hpge_peak

lower_range, upper_range, n_sig, mu, sigma, frac1, tau, n_bkg, hstep = range(9)
pars = [np.inf, np.inf, 1, 0, 1, 0, 0.1, 0, 0]
cov = [
    [1e-16, 0, 0, 0, 0, 0, 0, 0, 0],  # dxlo2
    [0, 1e-16, 0, 0, 0, 0, 0, 0, 0],  # dxhi2
    [0, 0, 1e-16, 0, 0, 0, 0, 0, 0],  # damp2
    [0, 0, 0, 1e-16, 0, 0, 0, 0, 0],  # dmu2
    [0, 0, 0, 0, 1e-02, 0, 0, 0, 0],  # dsig2
    [0, 0, 0, 0, 0, 1e-16, 0, 0, 0],  # dhtail2
    [0, 0, 0, 0, 0, 0, 1e-16, 0, 0],  # dtau2
    [0, 0, 0, 0, 0, 0, 0, 1e-16, 0],  # dbg02
    [0, 0, 0, 0, 0, 0, 0, 0, 1e-16],  # dhs2
]


def test_get_mu():
    mu, mu_err = hpge_peak.get_mu(pars, cov)
    assert mu == 0
    assert mu_err == np.sqrt(1e-16)


def test_get_fwhm():
    fwhm, dfwhm = hpge_peak.get_fwhm(pars, cov)

    assert fwhm == approx(2.3548, rel=1e-5)
    assert dfwhm == approx(2.3548e-1, rel=1e-5)


def test_get_total_events():
    total_events, total_event_err = hpge_peak.get_total_events(pars, cov)

    assert total_events == 1
    assert total_event_err == approx(np.sqrt(2 * 1e-32))



from pathlib import Path

import awkward as ak
import numpy as np
import pytest
from lgdo import Array, Table, VectorOfVectors, lh5

from pygama.evt import build_evt

config_dir = Path(__file__).parent / "configs"


@pytest.fixture(scope="module")
def files_config_nowrite(lgnd_test_data):
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    return {
        "tcm": (lgnd_test_data.get_path(tcm_path), "hardware_tcm_1"),
        "dsp": (lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")), "dsp", "ch{}"),
        "hit": (lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")), "hit", "ch{}"),
        "evt": (None, "evt"),
    }


@pytest.fixture(scope="module")
def files_config_write(lgnd_test_data, tmp_dir):
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    outfile = f"{tmp_dir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    return {
        "tcm": (lgnd_test_data.get_path(tcm_path), "hardware_tcm_1"),
        "dsp": (lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")), "dsp", "ch{}"),
        "hit": (lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")), "hit", "ch{}"),
        "evt": (outfile, "evt"),
    }


def test_basics(files_config_nowrite):
    evt = build_evt(
        files_config_nowrite,
        config=f"{config_dir}/basic-evt-config.yaml",
    )

    assert "statement" in evt.multiplicity.attrs
    assert evt.multiplicity.attrs["statement"] == "0bb decay is real"
    assert sorted(evt.keys()) == [
        "aoe",
        "energy",
        "energy_all_above1MeV",
        "energy_any_above1MeV",
        "energy_hit_idx",
        "energy_id",
        "energy_idx",
        "energy_sum",
        "is_aoe_rejected",
        "is_usable_aoe",
        "multiplicity",
        "timestamp",
    ]

    ak_evt = evt.view_as("ak")

    assert ak.all(ak_evt.energy_sum == ak.sum(ak_evt.energy, axis=-1))

    eid = evt.energy_id.view_as("np")
    eidx = evt.energy_idx.view_as("np")
    eidx = eidx[eidx != 999999999999]

    ids = lh5.read(
        "hardware_tcm_1/table_key", files_config_nowrite["tcm"][0]
    ).flattened_data.nda
    ids = ids[eidx]
    assert ak.all(ids == eid[eid != 0])

    ehidx = evt.energy_hit_idx.view_as("np")
    ids = lh5.read(
        "hardware_tcm_1/row_in_table", files_config_nowrite["tcm"][0]
    ).flattened_data.nda
    ids = ids[eidx]
    assert ak.all(ids == ehidx[ehidx != 999999999999])


def test_field_nesting(files_config_nowrite):
    config = {
        "channels": {"geds_on": ["ch1084803", "ch1084804", "ch1121600"]},
        "outputs": [
            "sub1___timestamp",
            "sub2___multiplicity",
            "sub2___dummy",
        ],
        "operations": {
            "sub1___timestamp": {
                "channels": "geds_on",
                "aggregation_mode": "first_at:dsp.tp_0_est",
                "expression": "dsp.timestamp",
            },
            "sub2___multiplicity": {
                "channels": "geds_on",
                "aggregation_mode": "sum",
                "expression": "hit.cuspEmax_ctc_cal > 25",
                "initial": 0,
            },
            "sub2___dummy": {
                "channels": "geds_on",
                "aggregation_mode": "sum",
                "expression": "hit.cuspEmax_ctc_cal > evt.sub1___timestamp",
                "initial": 0,
            },
        },
    }

    evt = build_evt(
        files_config_nowrite,
        config=config,
    )

    assert isinstance(evt, Table)
    assert isinstance(evt.sub1, Table)
    assert isinstance(evt.sub2, Table)
    assert isinstance(evt.sub1.timestamp, Array)

    assert sorted(evt.keys()) == ["sub1", "sub2"]
    assert sorted(evt.sub1.keys()) == ["timestamp"]
    assert sorted(evt.sub2.keys()) == ["dummy", "multiplicity"]


# FIXME: this can't be properly tested until proper testdata is available
# def test_spms_module(lgnd_test_data, files_config_nowrite):
#     build_evt(
#         files_config_nowrite,
#         config=f"{config_dir}/spms-module-config.yaml",
#         wo_mode="of",
#     )

#     outfile = files_config_nowrite["evt"][0]

#     evt = lh5.read("/evt", outfile)

#     t0 = ak.fill_none(ak.nan_to_none(evt.t0.view_as("ak")), 48_000)
#     tr_pos = evt.trigger_pos.view_as("ak") * 16
#     assert ak.all(tr_pos > t0 - 30_000)
#     assert ak.all(tr_pos < t0 + 30_000)

#     mask = evt._pulse_mask
#     assert isinstance(mask, VectorOfVectors)
#     assert len(mask) == 10
#     assert mask.ndim == 3

#     full = evt.spms_amp_full.view_as("ak")
#     amp = evt.spms_amp.view_as("ak")
#     assert ak.all(amp > 0.1)

#     assert ak.all(full[mask.view_as("ak")] == amp)

#     wo_empty = evt.spms_amp_wo_empty.view_as("ak")
#     assert ak.all(wo_empty == amp[ak.count(amp, axis=-1) > 0])

#     rawids = evt.rawid.view_as("ak")
#     assert rawids.ndim == 2
#     assert ak.count(rawids) == 30

#     idx = evt.hit_idx.view_as("ak")
#     assert idx.ndim == 2
#     assert ak.count(idx) == 30

#     rawids_wo_empty = evt.rawid_wo_empty.view_as("ak")
#     assert ak.count(rawids_wo_empty) == 7

#     vhit = evt.is_valid_hit.view_as("ak")
#     vhit.show()
#     assert ak.all(ak.num(vhit, axis=-1) == ak.num(full, axis=-1))


def test_vov(files_config_nowrite):
    evt = build_evt(
        files_config_nowrite,
        config=f"{config_dir}/vov-test-evt-config.json",
    )

    f_tcm = files_config_nowrite["tcm"][0]

    assert len(evt.keys()) == 12

    assert np.all(~np.isnan(evt.timestamp.nda))

    assert isinstance(evt.energy, VectorOfVectors)
    assert isinstance(evt.aoe, VectorOfVectors)
    assert isinstance(evt.multiplicity, Array)
    assert isinstance(evt.energy_times_aoe, VectorOfVectors)
    assert isinstance(evt.energy_times_multiplicity, VectorOfVectors)
    assert isinstance(evt.multiplicity_squared, Array)

    assert evt.energy.dtype == "float32"
    assert evt.aoe.dtype == "float64"
    assert evt.multiplicity.dtype == "int16"

    assert (
        np.diff(evt.energy.cumulative_length.nda, prepend=[0]) == evt.multiplicity.nda
    ).all()

    ids = lh5.read_as("hardware_tcm_1/table_key", f_tcm, library="ak")
    ids = ak.unflatten(
        ak.flatten(ids)[ak.flatten(evt.energy_idx.view_as("ak"))],
        ak.count(evt.energy_idx.view_as("ak"), axis=-1),
    )
    assert ak.all(ids == evt.energy_id.view_as("ak"))

    assert ak.all(
        ak.isclose(
            evt.energy_sum.view_as("ak"),
            ak.nansum(evt.energy.view_as("ak"), axis=-1),
            rtol=1e-3,
        )
    )
    assert ak.all(evt.aoe.view_as("ak") == evt.aoe_idx.view_as("ak"))


def test_graceful_crashing(files_config_nowrite):
    with pytest.raises(TypeError):
        build_evt(files_config_nowrite, None)

    conf = {"operations": {}}
    with pytest.raises(ValueError):
        build_evt(files_config_nowrite, conf)

    conf = {"channels": {"geds_on": ["ch1084803", "ch1084804", "ch1121600"]}}
    with pytest.raises(ValueError):
        build_evt(files_config_nowrite, conf)

    conf = {
        "channels": {"geds_on": ["ch1084803", "ch1084804", "ch1121600"]},
        "outputs": ["foo"],
        "operations": {
            "foo": {
                "channels": "geds_on",
                "aggregation_mode": "banana",
                "expression": "hit.cuspEmax_ctc_cal > a",
                "parameters": {"a": 25},
                "initial": 0,
            }
        },
    }
    with pytest.raises(ValueError):
        build_evt(
            files_config_nowrite,
            conf,
        )


def test_query(files_config_nowrite):
    evt = build_evt(
        files_config_nowrite,
        config=f"{config_dir}/query-test-evt-config.json",
    )

    assert len(evt.keys()) == 12


def test_vector_sort(files_config_nowrite):
    conf = {
        "channels": {"geds_on": ["ch1084803", "ch1084804", "ch1121600"]},
        "outputs": ["acend_id", "t0_acend", "decend_id", "t0_decend"],
        "operations": {
            "acend_id": {
                "channels": "geds_on",
                "aggregation_mode": "gather",
                "query": "hit.cuspEmax_ctc_cal>25",
                "expression": "tcm.table_key",
                "sort": "ascend_by:dsp.tp_0_est",
            },
            "t0_acend": {
                "aggregation_mode": "keep_at_ch:evt.acend_id",
                "expression": "dsp.tp_0_est",
            },
            "decend_id": {
                "channels": "geds_on",
                "aggregation_mode": "gather",
                "query": "hit.cuspEmax_ctc_cal>25",
                "expression": "tcm.table_key",
                "sort": "descend_by:dsp.tp_0_est",
            },
            "t0_decend": {
                "aggregation_mode": "keep_at_ch:evt.acend_id",
                "expression": "dsp.tp_0_est",
            },
        },
    }

    evt = build_evt(
        files_config_nowrite,
        conf,
    )

    assert len(evt.keys()) == 4
    vov_t0 = evt.t0_acend
    nda_t0 = vov_t0.to_aoesa().view_as("np")
    assert ((np.diff(nda_t0) >= 0) | (np.isnan(np.diff(nda_t0)))).all()
    vov_t0 = evt.t0_decend
    nda_t0 = vov_t0.to_aoesa().view_as("np")
    assert ((np.diff(nda_t0) <= 0) | (np.isnan(np.diff(nda_t0)))).all()


def test_build_evt_write(files_config_write):
    build_evt(
        files_config_write,
        config=f"{config_dir}/basic-evt-config.yaml",
        wo_mode="of",
    )
    outfile = files_config_write["evt"][0]
    assert Path(outfile).exists()



import lgdo.lh5 as lh5
import numpy as np

import pygama.pargen.AoE_cal as Coe


def test_aoe_cal(lgnd_test_data):
    # load test data here
    data = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/dsp/cal/p03/r000/l200-p03-r000-cal-20230311T235840Z-tier_dsp.lh5"
    )

    df = lh5.read_as("ch1104000/dsp", data, "pd")

    df["AoE_Uncorr"] = df["A_max"] / df["cuspEmax"]

    df["cuspEmax_cal"] = df["cuspEmax"] * 0.155

    cal_dict = {
        "AoE_Uncorr": {
            "expression": "A_max/cuspEmax",
            "parameters": {},
        }
    }

    aoe = Coe.CalAoE(
        cal_dict,
        "cuspEmax_cal",
        lambda x: np.sqrt(1.5 + 0.1 * x),
        selection_string="index==index",
        debug_mode=True,
    )
    aoe.calibrate(df, "AoE_Uncorr")
    assert (
        (aoe.low_cut_val < -1.0) & (aoe.low_cut_val > -3) & (~np.isnan(aoe.low_cut_val))
    )
    assert (
        (~np.isnan(aoe.low_side_sfs.loc[2614.5]["sf"]))
        & (aoe.low_side_sfs.loc[2614.5]["sf"] < 20)
        & (aoe.low_side_sfs.loc[2614.5]["sf"] > 0)
    )



"""
Module for cross talk correction of energies.
"""

from __future__ import annotations

import awkward as ak
import numpy as np
from dbetto import Props
from lgdo import lh5, types
from numpy.typing import ArrayLike

from pygama.hit.build_hit import _remove_uneeded_operations, _reorder_table_operations

from .. import utils


def build_tcm_index_array(
    tcm: utils.TCMData, datainfo: utils.DataInfo, rawids: np.ndarray
) -> np.ndarray:
    """Builds a TCM index array for use in the event tier.

    Parameters
    ----------
    datainfo
        :class:`.DataInfo` object.
    tcm
        time-coincidence map object.
    rawids
        list of channel rawids from the cross talk matrix.
    """

    # initialise the output object
    tcm_indexs_out = np.full((len(tcm.table_key), len(rawids)), np.nan)

    # parse observables string. default to hit tier
    for idx_chan, channel in enumerate(rawids):

        # get the event indexes
        table_id = utils.get_tcm_id_by_pattern(
            datainfo._asdict()["dsp"].table_fmt, f"ch{channel}"
        )
        if table_id is None:
            continue
        chan_tcm_indexs = np.where(ak.flatten(tcm.table_key) == table_id)[0].to_numpy()
        tbl_idxs_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()
        tcm_indexs_out[tbl_idxs_ch, idx_chan] = chan_tcm_indexs

    # transpose to return object where row is events and column rawid idx
    return tcm_indexs_out


def gather_energy(
    observable: str, tcm: utils.TCMData, datainfo: utils.DataInfo, rawids: ArrayLike
) -> ArrayLike:
    """Prepares the array of energies for the cross talk correction.

    Parameters
    ----------
    observable
        expression for the pulse parameter to be gathered, can be a combination
        of different fields.
    datainfo
        :class:`.DataInfo` object
    tcm
        time-coincidence map object.
    rawids
        list of channel rawids from the cross talk matrix.
    """

    # replace group. with group___
    for tier in datainfo._asdict():
        group = datainfo._asdict()[tier].group
        observable = observable.replace(f"{group}.", f"{group}___")

    observable = observable.replace(".", "__")

    c = compile(observable, "gcc -O3 -ffast-math build_hit.py", "eval")

    tier_params = []
    for name in c.co_names:
        if "___" in name:
            tier, column = name.split("___")
            group = datainfo._asdict()[tier].group
            file = datainfo._asdict()[tier].file
            if (name, file, group, column) not in tier_params:
                tier_params.append((name, file, group, column))

    # initialise the output object
    energy_out = np.full((len(tcm.table_key), len(rawids)), np.nan)

    for idx_chan, channel in enumerate(rawids):
        tbl = types.Table()
        chan_tcm_indexs = np.where(ak.flatten(tcm.table_key) == channel)[0].to_numpy()
        tbl_idxs_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()
        evt_ids_ch = np.repeat(
            np.arange(0, len(tcm.table_key)), ak.sum(tcm.table_key == channel, axis=1)
        )

        for name, file, group, column in tier_params:
            try:
                # read the energy data
                data = lh5.read(f"ch{channel}/{group}/{column}", file, idx=tbl_idxs_ch)
                tbl.add_column(name, data)
            except (lh5.exceptions.LH5DecodeError, KeyError):
                tbl.add_column(name, types.Array(np.full_like(evt_ids_ch, np.nan)))

        res = tbl.eval(observable)
        energy_out[evt_ids_ch, idx_chan] = res.nda

    return energy_out


def filter_hits(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    filter_expr: str,
    xtalk_corr_energy: ArrayLike,
    rawids: ArrayLike,
) -> ArrayLike:
    """Function that removes hits in an event below threshold.

    Parameters:
    -----------
    datainfo, tcm
        :class:`.DataInfo` and :class:`.TCMData` objects.
    filter_expr
        string containing the logic used to define which events are above
        threshold.  this string can also refer to the corrected energy as
        `xtalk_corr_energy`.
    xtalk_corr_energy
        2D numpy array of correct energy, the row corresponds to the event and
        the column the rawid.
    rawids
        1D array of the rawids corresponding to each column.
    """

    # find the fields in the string
    mask = np.full_like(xtalk_corr_energy, False, dtype=bool)

    # replace group. with group___
    for tier in datainfo._asdict():
        group = datainfo._asdict()[tier].group
        filter_expr = filter_expr.replace(f"{group}.", f"{group}___")

    c = compile(filter_expr, "toby dixon was here", "eval")

    tier_params = []
    for name in c.co_names:
        if "___" in name:
            tier, column = name.split("___")
            group = datainfo._asdict()[tier].group
            file = datainfo._asdict()[tier].file
            if (name, file, group, column) not in tier_params:
                tier_params.append((name, file, group, column))

    for idx_chan, channel in enumerate(rawids):
        tbl = types.Table()

        chan_tcm_indexs = np.where(ak.flatten(tcm.table_key) == channel)[0].to_numpy()
        tbl_idxs_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()
        evt_ids_ch = np.repeat(
            np.arange(0, len(tcm.table_key)), ak.sum(tcm.table_key == channel, axis=1)
        )

        for name, file, group, column in tier_params:
            try:
                # read the energy data
                data = lh5.read(f"ch{channel}/{group}/{column}", file, idx=tbl_idxs_ch)

                tbl.add_column(name, data)
            except (lh5.exceptions.LH5DecodeError, KeyError):
                tbl.add_column(name, types.Array(np.full_like(evt_ids_ch, np.nan)))

        # add the corrected energy to the table
        tbl.add_column(
            "xtalk_corr_energy", types.Array(xtalk_corr_energy[evt_ids_ch, idx_chan])
        )
        res = tbl.eval(filter_expr)
        mask[evt_ids_ch, idx_chan] = res.nda

    return mask


def xtalk_correct_energy_impl(
    uncal_energy: ArrayLike,
    cal_energy: ArrayLike,
    xtalk_matrix: ArrayLike,
    xtalk_threshold: float = None,
):
    r"""Function to perform the actual xtalk correction of energy.

    1. The energies are converted to a sparse format where each row corresponds
       to a rawid
    2. All energy less than the threshold are set to 0
    3. The correction is computed as:

    .. math::

       E_{\text{cor},i} = -\times M_{i,j}E_{j}

    where $M_{i,j}$ is the cross talk matrix element where $i$ is response and
    $j$ trigger channel.

    Parameters
    ----------
    uncal_energy
        2D numpy array of the uncalibrated energies in each event, the row
        corresponds to an event and the column the rawid.
    cal_energy
        2D numpy array of the calibrated energies in each event, the row
        corresponds to an event and the column the rawid.
    xtalk_matrix
        2D numpy array of the cross talk correction matrix, the indices should
        correspond to rawids (with same mapping as energies).
    xtalk_threshold
        threshold below which a hit is not used in xtalk correction.

    """
    # check input shapes and sizes
    uncal_energy_no_nan = np.nan_to_num(uncal_energy, 0)
    cal_energy_no_nan = np.nan_to_num(cal_energy, 0)

    if xtalk_threshold is not None:
        uncal_energy_with_threshold = np.where(
            cal_energy_no_nan < xtalk_threshold, 0, uncal_energy_no_nan
        )
    else:
        uncal_energy_with_threshold = uncal_energy_no_nan
    energy_correction = -np.matmul(xtalk_matrix, uncal_energy_with_threshold.T).T
    return uncal_energy_no_nan + energy_correction


def get_xtalk_correction(
    tcm: utils.DataInfo,
    datainfo: utils.DataInfo,
    uncal_energy_expr: str,
    cal_energy_expr: str,
    xtalk_threshold: float = None,
    xtalk_matrix_filename: str = "",
    xtalk_rawid_obj: str = "xtc/rawid_index",
    xtalk_matrix_obj: str = "xtc/xtalk_matrix_negative",
    positive_xtalk_matrix_obj: str = "xtc/xtalk_matrix_positive",
):

    # read lh5 files to numpy
    xtalk_matrix_numpy = lh5.read_as(xtalk_matrix_obj, xtalk_matrix_filename, "np")
    xtalk_matrix_rawids = lh5.read_as(xtalk_rawid_obj, xtalk_matrix_filename, "np")

    positive_xtalk_matrix_numpy = lh5.read_as(
        positive_xtalk_matrix_obj, xtalk_matrix_filename, "np"
    )

    # Combine positive and negative matrixs
    # Now the matrix should have negative values corresponding to negative cross talk
    # and positive values corresponding to positive cross talk .
    # we also set nan to 0 and we transpose so that the row corresponds to response and column trigger
    xtalk_matrix = np.nan_to_num(
        np.where(
            abs(xtalk_matrix_numpy) > abs(positive_xtalk_matrix_numpy),
            xtalk_matrix_numpy,
            positive_xtalk_matrix_numpy,
        ),
        0,
    ).T

    uncal_energy_array = gather_energy(
        uncal_energy_expr, tcm, datainfo, xtalk_matrix_rawids
    )
    cal_energy_array = gather_energy(
        cal_energy_expr, tcm, datainfo, xtalk_matrix_rawids
    )

    energy_corr = xtalk_correct_energy_impl(
        uncal_energy_array, cal_energy_array, xtalk_matrix, xtalk_threshold
    )
    return energy_corr


def calibrate_energy(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    energy_corr: np.ndarray,
    xtalk_matrix_rawids: np.ndarray,
    par_files: str | list[str],
    uncal_energy_var: str = None,
    recal_energy_var: str = None,
    channel_mapping: dict = None,
):
    """Function to recalibrate the energy after xtalk correction.

    Parameters
    ---------
    datainfo
        :class:`.DataInfo` object.
    tcm
        :class:`.TCMData` object.
    energy_corr
        cross talk corrected (uncal) energies to be recalibrated.
    par_files
        path to the parameter files.
    uncal_energy_var
        name of the uncalibrated energy variable.
    recal_energy_var
        variable to be used for recalibration.
    """

    out_arr = np.full_like(energy_corr, np.nan)
    par_dicts = Props.read_from(par_files)
    pars = {
        chan: chan_dict["pars"]["operations"] for chan, chan_dict in par_dicts.items()
    }

    p = uncal_energy_var.split(".")
    tier = p[0] if len(p) > 1 else "dsp"

    table_fmt = datainfo._asdict()[tier].table_fmt
    file = datainfo._asdict()[tier].file

    for i, chan in enumerate(xtalk_matrix_rawids):
        try:
            if f"ch{chan}" in pars:
                cfg = pars[f"ch{chan}"]
            elif channel_mapping is not None and f"ch{chan}" in channel_mapping:
                cfg = pars[channel_mapping[f"ch{chan}"]]
            else:
                raise KeyError
            cfg, chan_inputs = _remove_uneeded_operations(
                _reorder_table_operations(cfg), recal_energy_var.split(".")[-1]
            )

            chan_inputs.remove(uncal_energy_var.split(".")[-1])

            # get the event indices
            table_id = utils.get_tcm_id_by_pattern(table_fmt, f"ch{chan}")
            if table_id is None:
                continue

            chan_tcm_indexs = np.where(ak.flatten(tcm.table_key) == table_id)[
                0
            ].to_numpy()
            tbl_idxs_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()
            evt_ids_ch = ak.any(tcm.table_key == table_id, axis=1)

            # read the dsp data
            outtbl_obj = lh5.read(
                f"ch{chan}/dsp/", file, idx=tbl_idxs_ch, field_mask=chan_inputs
            )

            # add the uncalibrated energy to the table
            outtbl_obj.add_column(
                uncal_energy_var.split(".")[-1],
                types.Array(energy_corr[evt_ids_ch, i]),
            )

            for outname, info in cfg.items():
                outcol = outtbl_obj.eval(
                    info["expression"], info.get("parameters", None)
                )
                outtbl_obj.add_column(outname, outcol)
            out_arr[evt_ids_ch, i] = outtbl_obj[recal_energy_var.split(".")[-1]].nda
        except KeyError:
            out_arr[:, i] = np.nan

    return out_arr



"""
Utilities for grouping hit data into events.
"""

from .build_evt import build_evt
from .build_tcm import build_tcm
from .tcm import generate_tcm_cols

__all__ = ["build_tcm", "generate_tcm_cols", "build_evt"]



r"""
Provide a convenience function for a Gaussian on top of an exponential. The correct usage is :func:`gauss_on_exponential.pdf_ext(x, n_sig, mu, sigma, lambda, n_bkg, mu_exp, sigma_exp)`

Parameters
----------
n_sig
    The area of the Gaussian
mu, sigma
    The location and scale of the first Gaussian
n_bkg
    The area of the exponential
lambda
    The characteristic scale of the exponential
mu_exp, sigma_exp
    The location and scale of the exponential

Returns
-------
gauss_on_exponential
    A subclass of :class:`SumDists` and :class:`rv_continuous`, has methods of :func:`pdf`, :func:`cdf`, etc.
"""

from pygama.math.functions.exponential import exponential
from pygama.math.functions.gauss import gaussian
from pygama.math.functions.sum_dists import SumDists

(n_sig, mu, sigma, n_bkg, lamb, mu_exp, sigma_exp) = range(
    7
)  # the sum_dist array we will pass will be n_sig, mu, sigma, lambda, n_bkg, mu_exp, sigma_exp
par_array = [(gaussian, [mu, sigma]), (exponential, [mu_exp, sigma_exp, lamb])]

gauss_on_exponential = SumDists(
    par_array,
    [n_sig, n_bkg],
    "areas",
    parameter_names=["n_sig", "mu", "sigma", "n_bkg", "lambd", "mu_exp", "sigma_exp"],
    name="gauss_on_exponential",
)



from pygama.evt import utils


def test_tier_data_tuple():
    files = utils.make_files_config(
        {
            "tcm": ("f1", "g1"),
            "dsp": ("f2", "g2"),
            "hit": ("f3", "g3"),
            "evt": ("f4", "g4"),
        }
    )

    assert files.raw == utils.H5DataLoc()
    assert files.tcm.file == "f1"
    assert files.tcm.group == "g1"
    assert files.dsp.file == "f2"
    assert files.dsp.group == "g2"
    assert files.hit.file == "f3"
    assert files.hit.group == "g3"
    assert files.evt.file == "f4"
    assert files.evt.group == "g4"



from __future__ import annotations

import logging
from collections import namedtuple

import numpy as np
import pandas as pd
from lgdo.types import Table, VectorOfVectors

# later on we might want a tcm class, or interface / inherit from an entry list
# class. For now we just need the key clustering functionality

log = logging.getLogger(__name__)

coin_groups = namedtuple("coin_groups", ["name", "window", "window_ref"])


def generate_tcm_cols(
    iterators: list,
    coin_windows: float = 0,
    table_keys: list[int] = None,
    row_in_tables: list[int] = None,
    fields: list[str] = None,
) -> dict[np.ndarray]:
    r"""Generate the columns of a time coincidence map.

    Generate the columns of a time coincidence map from a list of arrays of
    coincidence data (e.g. hit times from different channels). Returns 3
    :class:`numpy.ndarray`\ s representing a vector-of-vector-like structure:
    two flattened arrays ``table_key`` (e.g. channel number) and  ``row_in_table``
    (e.g. hit index) that specify the location in the input ``coin_data`` of
    each datum belonging to a coincidence event, and a ``cumulative_length``
    array that specifies which rows of the other two output arrays correspond
    to which coincidence event. These can be used to retrieve other data at
    the same tier as the input data into coincidence structures.

    The 0'th entry of ``cumulative_length`` contains the number of hits in the
    zeroth coincidence event, and the i'th entry is set to
    ``cumulative_length[i-1]`` plus the number of hits in the i'th event.
    Thus, the hits of the i'th event can be found in rows
    ``cumulative_length[i-1]`` to ``cumulative_length[i] - 1`` of ``table_key``
    and ``row_in_table``.

    An example: ``cumulative_length = [4, 7, ...]``.  Then rows 0 to 3 in
    `table_key` and `row_in_table` correspond to the hits in event 0, rows 4 to 6
    correspond to event 1, and so on.

    Makes use of :func:`pandas.concat`, :meth:`pandas.DataFrame.sort_values`,
        and :meth:`pandas.DataFrame.diff` functions:

    - pull data into a :class:`pandas.DataFrame`
    - sort events by strictly ascending value of `coin_col`
    - group hits if the difference in `coin_data` is less than `coin_window`

    Parameters
    ----------
    coin_data
        a list of arrays of the data to be clustered.
    coin_window
        the clustering window width. `coin_data` within the `coin_window` get
        aggregated into the same coincidence cluster. A value of ``0`` means an
        equality test.
    window_ref
        when testing one datum for inclusion in a cluster, test if it is within
        `coin_window` of

        - ``"first"`` -- the first element in the cluster (rigid window width)
        - ``"last"`` -- the last element in the cluster (window grows until two
          data are separated by more than coin_window)

    table_keys
        if provided, use `table_keys` in place of "index in coin_data" as the
        integer corresponding to each element of `coin_data` (e.g. a channel
        number).
    row_in_tables
        if provided, use these values in places of the ``DataFrame`` index for
        the return values of `row_in_table`.

    Returns
    -------
    col_dict
        keys are ``cumulative_length``, ``table_key``, and  ``row_in_table``.
        ``cumulative_length`` specifies which rows of the other two output
        arrays correspond to which coincidence event. ``table_key`` and
        ``row_in_table`` specify the location in ``coin_data`` of each datum
        belonging to the coincidence event.
    """
    if isinstance(iterators, list):
        iterators = np.array(iterators)

    tcm = None
    at_end = np.zeros(len(iterators), dtype=bool)
    skip_mask = np.zeros(len(iterators), dtype=bool)
    buffer = None
    if table_keys is None:
        table_keys = np.arange(0, len(iterators))
    while not at_end.all():
        curr_mask = ~skip_mask & ~at_end
        dfs = []
        for _ii, it in enumerate(iterators[curr_mask]):
            ii = np.where(curr_mask)[0][_ii]
            try:
                buffer = it.__next__()
                buf_len = len(buffer)
                start = it.current_i_entry
            except StopIteration:
                at_end[ii] = True
                continue
            if buf_len < it.buffer_len:
                at_end[ii] = True
            table_key = table_keys[ii]
            table_key = np.full(buf_len, table_key, dtype=int)
            buffer = buffer.view_as("pd")[:buf_len]
            buffer["table_key"] = table_key
            if row_in_tables is not None:
                buffer["row_in_table"] = row_in_tables.astype(int)[ii][
                    start : start + buf_len
                ]
            else:
                buffer["row_in_table"] = np.arange(start, start + buf_len, dtype=int)
            if len(buffer) > 0:
                dfs.append(buffer)  # don't copy the data!

        if at_end.all() and len(dfs) == 0 and tcm is None:
            break

        if tcm is None:
            tcm = pd.concat(dfs).sort_values(
                [entry.name for entry in coin_windows] + ["table_key"]
            )
        else:
            tcm = pd.concat([tcm] + dfs).sort_values(
                [entry.name for entry in coin_windows] + ["table_key"]
            )

        # define mask, true when new event, false if part of same event
        mask = np.zeros(len(tcm) - 1, dtype=bool)
        for entry in coin_windows:
            diffs = np.diff(tcm[entry.name])
            if entry.window_ref == "last":
                mask = mask | (diffs > entry.window)
            else:
                raise NotImplementedError(f"window_ref {entry.window_ref}")

        # grab up to evt including last instance of a channel to know that all channels
        # have been included in previous evts
        last_instance = {
            arr_id: index for index, arr_id in enumerate(tcm.table_key.to_numpy())
        }
        log.debug(f"last instance: {last_instance}")

        for i, entry in enumerate(table_keys):
            if entry not in last_instance:
                last_instance[entry] = np.inf
            if at_end[i]:
                last_instance[entry] = np.inf

        if len(np.array(table_keys)[~at_end]) > 1:
            comp_chan = np.array(table_keys)[~at_end][0]
            skip_mask = np.array(
                [(last_instance[arr] >= last_instance[comp_chan]) for arr in table_keys]
            )
            if skip_mask.all():
                skip_mask = np.zeros(len(table_keys), dtype=bool)
        else:
            skip_mask = np.zeros(len(table_keys), dtype=bool)

        # want to write entries only up to last entry of a channel to ensure all included in evt
        if at_end.all():
            log.debug("at end, writing all entries")
            write_mask = mask
            last_entry = None
        else:
            last_instance = int(np.min([last_instance[arr] for arr in table_keys]))
            last_entry = np.where(mask[:last_instance])[0]
            log.debug(f"last instance: {last_instance}")
            log.debug(f"last entry: {last_entry}")

            if len(last_entry) == 0:
                log.debug("last entry 0, going to next iteration")
                log.debug(tcm)
                continue
            else:
                last_entry = last_entry[-1] + 1
                write_mask = mask[:last_entry]
                if len(write_mask) == 0:
                    log.debug("no entries, going to next iteration")
                    log.debug(tcm)
                    continue

        # get cumulative_length
        cumulative_length = np.array(np.where(write_mask)[0]) + 1
        if at_end.all():
            cumulative_length = np.append(cumulative_length, len(write_mask) + 1)

        out_tbl = Table(size=len(cumulative_length))
        out_tbl.add_field(
            "table_key",
            VectorOfVectors(
                cumulative_length=cumulative_length,
                flattened_data=tcm["table_key"].to_numpy()[:last_entry],
            ),
        )
        out_tbl.add_field(
            "row_in_table",
            VectorOfVectors(
                cumulative_length=cumulative_length,
                flattened_data=tcm["row_in_table"].to_numpy()[:last_entry],
            ),
        )
        if fields is not None:
            for f in fields:
                out_tbl.add_field(
                    f,
                    VectorOfVectors(
                        cumulative_length=cumulative_length,
                        flattened_data=tcm[f].to_numpy()[:last_entry],
                    ),
                )
        if last_entry is None:
            tcm = None
        else:
            tcm = tcm[last_entry:]
        yield out_tbl



r"""
Contains a list of distribution functions, all implemented using Numba's
:func:`numba.jit` to take advantage of the just-in-time speed boost.
"""

# nopycln: file

from pygama.math.functions.crystal_ball import crystal_ball  # noqa: F401
from pygama.math.functions.crystal_ball import (  # noqa: F401
    nb_crystal_ball_cdf,
    nb_crystal_ball_pdf,
)
from pygama.math.functions.error_function import nb_erf, nb_erfc  # noqa: F401
from pygama.math.functions.exgauss import exgauss  # noqa: F401
from pygama.math.functions.exgauss import (  # noqa: F401
    nb_exgauss_cdf,
    nb_exgauss_pdf,
    nb_gauss_tail_approx,
)
from pygama.math.functions.exponential import exponential  # noqa: F401
from pygama.math.functions.gauss import (  # noqa: F401
    gaussian,
    nb_gauss,
    nb_gauss_cdf,
    nb_gauss_pdf,
    nb_gauss_scaled_pdf,
)
from pygama.math.functions.gauss_on_exgauss import gauss_on_exgauss  # noqa: F401
from pygama.math.functions.gauss_on_linear import gauss_on_linear  # noqa: F401
from pygama.math.functions.gauss_on_step import gauss_on_step  # noqa: F401
from pygama.math.functions.gauss_on_uniform import gauss_on_uniform  # noqa: F401
from pygama.math.functions.hpge_peak import hpge_peak  # noqa: F401
from pygama.math.functions.linear import linear  # noqa: F401
from pygama.math.functions.moyal import moyal  # noqa: F401
from pygama.math.functions.polynomial import nb_poly  # noqa: F401
from pygama.math.functions.step import (  # noqa: F401
    nb_step_cdf,
    nb_step_pdf,
    nb_unnorm_step_pdf,
    step,
)
from pygama.math.functions.triple_gauss_on_double_step import (  # noqa: F401
    triple_gauss_on_double_step,
)
from pygama.math.functions.uniform import uniform  # noqa: F401



import numpy as np
from scipy.stats import norm

from pygama.math.functions.gauss import gaussian
from pygama.math.functions.pygama_continuous import NumbaFrozen


def test_numba_frozen():
    x = np.arange(-10, 10)
    mu = 2
    sigma = 3
    gauss = gaussian(mu, sigma)
    assert isinstance(gauss, NumbaFrozen)

    numba_pdf = gauss.pdf(x)
    scipy_pdf = norm.pdf(x, mu, sigma)
    assert np.allclose(numba_pdf, scipy_pdf)

    numba_get_pdf = gauss.get_pdf(x)
    assert np.allclose(numba_get_pdf, scipy_pdf)

    numba_pdf_norm = gauss.pdf_norm(x, x[0], x[-1])
    scipy_pdf_norm = scipy_pdf / (np.diff(norm.cdf([x[0], x[-1]], mu, sigma)))
    assert np.allclose(numba_pdf_norm, scipy_pdf_norm)

    numba_cdf = gauss.cdf(x)
    scipy_cdf = norm.cdf(x, mu, sigma)
    assert np.allclose(numba_cdf, scipy_cdf)

    numba_get_cdf = gauss.get_cdf(x)
    assert np.allclose(numba_get_cdf, scipy_cdf)

    numba_cdf_norm = gauss.cdf_norm(x, x[0], x[-1])
    scipy_cdf_norm = scipy_cdf / (np.diff(norm.cdf([x[0], x[-1]], mu, sigma)))
    assert np.allclose(numba_cdf_norm, scipy_cdf_norm)

    area = 10
    x_lo = -20
    x_hi = 20

    numba_extended_pdf_area, numba_extended_pdf = gauss.pdf_ext(x, x_lo, x_hi, area)
    assert np.allclose(numba_extended_pdf_area, area, rtol=1e-3)
    assert np.allclose(numba_extended_pdf, area * scipy_pdf)

    numba_extended_cdf = gauss.cdf_ext(x, area)
    assert np.allclose(numba_extended_cdf, area * scipy_cdf)



"""
Exponential distributions for pygama
"""

import numba as nb
import numpy as np

from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.utils import numba_math_defaults as nb_defaults
from pygama.utils import numba_math_defaults_kwargs as nb_kwargs


@nb.njit(**nb_kwargs)
def nb_exponential_pdf(
    x: np.ndarray, mu: float, sigma: float, lamb: float
) -> np.ndarray:
    r"""
    Normalised exponential probability density distribution, w/ args: mu, sigma, lamb. Its range of support is :math:`x\in[0,\infty), \lambda>0`.
    It computes:


    .. math::
        pdf(x, \lambda, \mu, \sigma) = \begin{cases} \lambda e^{-\lambda\frac{x-\mu}{\sigma}} \quad , \frac{x-\mu}{\sigma}\geq 0 \\ 0 \quad , \frac{x-\mu}{\sigma}<0 \end{cases}


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.


    Parameters
    ----------
    x
        The input data
    mu
        The amount to shift the distribution
    sigma
        The amount to scale the distribution
    lamb
        The rate
    """

    y = np.empty_like(x, dtype=np.float64)
    for i in nb.prange(x.shape[0]):
        y[i] = (x[i] - mu) / sigma
        if y[i] < 0:
            y[i] = 0
        else:
            y[i] = (lamb * np.exp(-1 * lamb * y[i])) / sigma
    return y


@nb.njit(**nb_kwargs)
def nb_exponential_cdf(
    x: np.ndarray, mu: float, sigma: float, lamb: float
) -> np.ndarray:
    r"""
    Normalised exponential cumulative distribution, w/ args:  mu, sigma, lamb. Its range of support is :math:`x\in[0,\infty), \lambda>0`.
    It computes:


    .. math::
        cdf(x, \lambda, \mu, \sigma) = \begin{cases}  1-e^{-\lambda\frac{x-\mu}{\sigma}} \quad , \frac{x-\mu}{\sigma} > 0 \\ 0 \quad , \frac{x-\mu}{\sigma}\leq 0 \end{cases}


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.


    Parameters
    ----------
    x
        The input data
    mu
        The amount to shift the distribution
    sigma
        The amount to scale the distribution
    lamb
        The rate
    """

    y = np.empty_like(x, dtype=np.float64)
    for i in nb.prange(x.shape[0]):
        y[i] = (x[i] - mu) / sigma
        if y[i] <= 0:
            y[i] = 0
        else:
            y[i] = 1 - np.exp(-1 * lamb * y[i])
    return y


@nb.njit(**nb_defaults(parallel=False))
def nb_exponential_scaled_pdf(
    x: np.ndarray, area: float, mu: float, sigma: float, lamb: float
) -> np.ndarray:
    r"""
    Scaled exponential probability distribution, w/ args: area, mu, sigma, lambd.
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        Input data
    area
        The number of counts in the signal
    mu
        The amount to shift the distribution
    sigma
        The amount to scale the distribution
    lamb
        The rate
    """

    return area * nb_exponential_pdf(x, mu, sigma, lamb)


@nb.njit(**nb_defaults(parallel=False))
def nb_exponential_scaled_cdf(
    x: np.ndarray, area: float, mu: float, sigma: float, lamb: float
) -> np.ndarray:
    r"""
    Exponential cdf scaled by the area, used for extended binned fits
    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.


    Parameters
    ----------
    x
        Input data
    area
        The number of counts in the signal
    mu
        The amount to shift the distribution
    sigma
        The amount to scale the distribution
    lamb
        The rate
    """

    return area * nb_exponential_cdf(x, mu, sigma, lamb)


class ExponentialGen(PygamaContinuous):

    def __init__(self, *args, **kwargs):
        self.x_lo = 0
        self.x_hi = np.inf
        super().__init__(*args, **kwargs)

    def _pdf(self, x: np.ndarray, mu: float, sigma: float, lamb: float) -> np.ndarray:
        x.flags.writeable = True
        return nb_exponential_pdf(x, mu[0], sigma[0], lamb[0])

    def _cdf(self, x: np.ndarray, mu: float, sigma: float, lamb: float) -> np.ndarray:
        x.flags.writeable = True
        return nb_exponential_cdf(x, mu[0], sigma[0], lamb[0])

    def get_pdf(
        self, x: np.ndarray, mu: float, sigma: float, lamb: float
    ) -> np.ndarray:
        return nb_exponential_pdf(x, mu, sigma, lamb)

    def get_cdf(
        self, x: np.ndarray, mu: float, sigma: float, lamb: float
    ) -> np.ndarray:
        return nb_exponential_cdf(x, mu, sigma, lamb)

    # needed so that we can hack iminuit's introspection to function parameter names... unless
    def pdf_norm(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        lamb: float,
    ) -> np.ndarray:
        return self._pdf_norm(x, x_lo, x_hi, mu, sigma, lamb)

    def cdf_norm(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        lamb: float,
    ) -> np.ndarray:
        return self._cdf_norm(x, x_lo, x_hi, mu, sigma, lamb)

    def pdf_ext(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        area: float,
        mu: float,
        sigma: float,
        lamb: float,
    ) -> np.ndarray:
        return np.diff(
            nb_exponential_scaled_cdf(np.array([x_lo, x_hi]), area, mu, sigma, lamb)
        )[0], nb_exponential_scaled_pdf(x, area, mu, sigma, lamb)

    def cdf_ext(
        self, x: np.ndarray, area: float, mu: float, sigma: float, lamb: float
    ) -> np.ndarray:
        return nb_exponential_scaled_cdf(x, area, mu, sigma, lamb)

    def required_args(self) -> tuple[str, str, str]:
        return "mu", "sigma", "lambda"


exponential = ExponentialGen(a=0.0, name="exponential")



import numpy as np
from scipy.stats import norm

from pygama.math.functions.gauss_on_linear import gauss_on_linear
from pygama.math.functions.sum_dists import SumDists


def line(x, m, b):
    return m * x + b


def quad(x, x_lo, m, b):
    return m * (x**2 - x_lo**2) / 2 + b * (x - x_lo)


def test_gauss_on_linear_pdf():
    x = np.arange(-10, 10)
    mu = 1
    sigma = 2
    x_lo = np.amin(x)
    x_hi = np.amax(x)
    m = 2
    b = 3
    n_sig = 10
    n_bkg = 20

    pars = np.array([x_lo, x_hi, n_sig, mu, sigma, n_bkg, m, b], dtype=float)

    assert isinstance(gauss_on_linear, SumDists)

    y_direct = gauss_on_linear.get_pdf(x, *pars)

    normalization = m / 2 * (x_hi**2 - x_lo**2) + b * (x_hi - x_lo)
    scipy_linear = n_bkg * line(x, m, b) / normalization

    scipy_gauss = n_sig * norm.pdf(x, mu, sigma)

    scipy_y = scipy_gauss + scipy_linear

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    x_lo = -10
    x_hi = 9

    y_sig, y_ext = gauss_on_linear.pdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)
    assert np.allclose(y_sig, n_sig + n_bkg, rtol=1e-3)


def test_gauss_on_linear_cdf():
    x = np.arange(-10, 10)
    mu = 1
    sigma = 2
    x_lo = np.amin(x)
    x_hi = np.amax(x)
    m = 2
    b = 3
    n_sig = 10
    n_bkg = 20

    pars = np.array([x_lo, x_hi, n_sig, mu, sigma, n_bkg, m, b], dtype=float)

    assert isinstance(gauss_on_linear, SumDists)

    y_direct = gauss_on_linear.get_cdf(x, *pars)

    scipy_gauss = n_sig * norm.cdf(x, mu, sigma)
    normalization = m / 2 * (x_hi**2 - x_lo**2) + b * (x_hi - x_lo)
    scipy_linear = n_bkg * quad(x, x_lo, m, b) / normalization

    scipy_y = scipy_gauss + scipy_linear

    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    y_ext = gauss_on_linear.cdf_ext(x, *pars)
    assert np.allclose(y_ext, scipy_y, rtol=1e-8)


def test_required_args():
    names = gauss_on_linear.required_args()
    assert names[0] == "x_lo"
    assert names[1] == "x_hi"
    assert names[2] == "n_sig"
    assert names[3] == "mu"
    assert names[4] == "sigma"
    assert names[5] == "n_bkg"
    assert names[6] == "m"
    assert names[7] == "b"


def test_name():
    assert gauss_on_linear.name == "gauss_on_linear"



"""Event processors for SiPM data."""

from __future__ import annotations

from collections.abc import Sequence

import awkward as ak
import numpy as np
from lgdo import lh5, types

from .. import utils
from . import larveto


def gather_pulse_data(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    table_names: Sequence[str],
    channel_mapping: dict,
    *,
    observable: str,
    pulse_mask: types.VectorOfVectors = None,
    a_thr_pe: float = None,
    t_loc_ns: float = None,
    dt_range_ns: Sequence[float] = None,
    t_loc_default_ns: float = None,
    drop_empty: bool = True,
) -> types.VectorOfVectors:
    """Gathers SiPM pulse data into a 3D :class:`~lgdo.types.vectorofvectors.VectorOfVectors`.

    The returned data structure specifies the event in the first axis, the SiPM
    channel in the second and the pulse index in the last.

    Pulse data can be optionally masked with `pulse_mask` or a mask can be
    built on the fly from the `a_thr_pe`, `t_loc_ns`, `dt_range_ns`,
    `t_loc_default_ns` arguments (see :func:`make_pulse_data_mask`).

    If `pulse_mask`, `a_thr_pe`, `t_loc_ns`, `dt_range_ns`, `t_loc_default_ns`
    are all ``None``, no masking is applied and the full data set is returned.

    Parameters
    ----------
    datainfo, tcm, table_names
        positional arguments automatically supplied by :func:`.build_evt`.
    observable
        name of the pulse parameter to be gathered, optionally prefixed by tier
        name (e.g. ``hit.energy_in_pe``). If no tier is specified, it defaults
        to ``hit``.
    pulse_mask
        3D mask object used to filter out pulse data. See
        :func:`make_pulse_data_mask`.
    a_thr_pe
        amplitude threshold (in photoelectrons) used to build a pulse mask with
        :func:`make_pulse_data_mask`, if `pulse_mask` is ``None``. The output
        pulse data will be such that the pulse amplitude is above this value.
    t_loc_ns
        location of the time window in which pulses must sit. If a 1D array is
        provided, it is interpreted as a list of locations for each event (can
        be employed to e.g. provide the actual HPGe pulse position)
    dt_range_ns
        tuple with dimension of the time window in which pulses must sit
        relative to `t_loc_ns`. If, for example, `t_loc_ns` is 48000 ns and
        `dt_range_ns` is (-1000, 5000) ns, the resulting window will be (47000,
        53000) ns.
    t_loc_default_ns
        default value for `t_loc_ns`, in case the supplied value is
        :any:`numpy.nan`.
    drop_empty
        if ``True``, drop empty arrays at the last axis (the pulse axis), i.e.
        drop channels with no pulse data. The filtering is applied after the
        application of the mask.
    """
    # parse observables string. default to hit tier
    p = observable.split(".")
    tier = p[0] if len(p) > 1 else "hit"
    column = p[1] if len(p) > 1 else p[0]

    tierinfo = datainfo._asdict()[tier]

    # loop over selected table_names and load hit data
    concatme = []
    # number of channels per event
    evt_length = np.diff(
        np.insert(types.VectorOfVectors(tcm.table_key).cumulative_length, 0, 0)
    )

    # construct global event ID's
    glob_ids = np.repeat(np.arange(0, len(tcm.table_key), 1), evt_length)

    for channel in table_names:
        table_id = utils.get_tcm_id_by_pattern(tierinfo.table_fmt, channel)
        if table_id is None:
            continue

        # determine list of indices found in the TCM that we want to load for channel
        chan_tcm_indexs = np.where(ak.flatten(tcm.table_key) == table_id)[0].to_numpy()
        tbl_idxs_ch = ak.flatten(tcm.row_in_table)[chan_tcm_indexs].to_numpy()

        # read the data in
        lgdo_obj = lh5.read(
            f"/{channel}/{tierinfo.group}/{column}", tierinfo.file, idx=tbl_idxs_ch
        )
        data = lgdo_obj.view_as("np")

        # remove nans (this happens when SiPM data is stored as ArrayOfEqualSizedArrays)
        data = ak.drop_none(ak.nan_to_none(data))

        glob_ids_ch = glob_ids[
            chan_tcm_indexs
        ]  # global ID's where channel had a trigger

        # count number of hits per channel for global events with trigger in channel, else 0
        glob_ids_cts = np.zeros(len(tcm.table_key), dtype=int)
        glob_ids_cts[glob_ids_ch] = ak.count(data, axis=1)

        # insert empty row [] for global events with no trigger in channel
        # unflatten to the number of hits in channel otherwise
        data = ak.unflatten(ak.flatten(data), glob_ids_cts)

        # increase the dimensionality by one (events)
        data = ak.unflatten(data, np.full(data.layout.length, 1, dtype="uint8"))

        concatme.append(data)

    # concatenate along the event axes (i.e. gather table_names together)
    data = ak.concatenate(concatme, axis=1)

    # check if user wants to apply a mask
    if pulse_mask is None and any(
        [kwarg is not None for kwarg in (a_thr_pe, t_loc_ns, dt_range_ns)]
    ):
        # generate the time/amplitude mask from parameters
        pulse_mask = make_pulse_data_mask(
            datainfo,
            tcm,
            table_names,
            channel_mapping,
            a_thr_pe=a_thr_pe,
            t_loc_ns=t_loc_ns,
            dt_range_ns=dt_range_ns,
            t_loc_default_ns=t_loc_default_ns,
        )

    if pulse_mask is not None:
        if not isinstance(pulse_mask, ak.Array):
            pulse_mask = pulse_mask.view_as("ak")

        # apply the mask
        data = data[pulse_mask]

    # remove empty arrays = table_names with no pulses
    if drop_empty:
        data = data[ak.count(data, axis=-1) > 0]

    return types.VectorOfVectors(data, attrs=utils.copy_lgdo_attrs(lgdo_obj))


def gather_tcm_data(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    table_names: Sequence[str],
    channel_mapping: dict,
    *,
    tcm_field="id",
    pulse_mask=None,
    a_thr_pe=None,
    t_loc_ns=None,
    dt_range_ns=None,
    t_loc_default_ns=None,
    drop_empty=True,
) -> types.VectorOfVectors:
    """Gather TCM data into a 2D :class:`~lgdo.types.vectorofvectors.VectorOfVectors`.

    The returned data structure specifies the event on the first axis and the
    TCM data (`id` or `idx`) on the second. Can be used to filter out data from
    :func:`gather_pulse_data` based on SiPM channel provenance (`id`) or to
    load hit data from lower tiers (with `idx`).

    If `drop_empty` is ``True``, channel ids with no pulse data associated are
    removed.

    See :func:`gather_pulse_data` for documentation about the other function
    arguments.
    """

    # list user wanted table names
    table_ids = [
        utils.get_tcm_id_by_pattern(datainfo.hit.table_fmt, id)
        for id in table_names
        if utils.get_tcm_id_by_pattern(datainfo.hit.table_fmt, id) is not None
    ]
    # find them in tcm.id (we'll filter the rest out)
    tcm_id_padded = types.VectorOfVectors(tcm.table_key).to_aoesa().view_as("np")
    locs = np.isin(tcm_id_padded, table_ids)

    # select tcm field requested by the user
    data = tcm._asdict()[tcm_field]

    # apply mask
    # NOTE: need to cast to irregular axes, otherwise the masking result is
    # non-nested
    data = data[ak.from_regular(locs)]

    # check if user wants to apply a custom mask
    if drop_empty:
        if pulse_mask is None:
            # generate the time/amplitude mask from parameters
            # if all parameters are None, a dummy mask (the identity) will be made
            pulse_mask = make_pulse_data_mask(
                datainfo,
                tcm,
                table_names,
                channel_mapping,
                a_thr_pe=a_thr_pe,
                t_loc_ns=t_loc_ns,
                dt_range_ns=dt_range_ns,
                t_loc_default_ns=t_loc_default_ns,
            )

        if not isinstance(pulse_mask, ak.Array):
            pulse_mask = pulse_mask.view_as("ak")

        if pulse_mask.ndim != 3:
            msg = "pulse_mask must be 3D"
            raise ValueError(msg)

        # convert the 3D mask to a 2D mask (can be used to filter table_ids)
        pulse_mask = pulse_mask[ak.num(pulse_mask, axis=2) > 0]
        ch_mask = ak.sum(pulse_mask, axis=-1) > 0

        # apply the mask
        data = data[ch_mask]

    return types.VectorOfVectors(data)


# NOTE: the mask never gets the empty arrays removed
def make_pulse_data_mask(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    table_names: Sequence[str],
    channel_mapping: dict,
    *,
    a_thr_pe=None,
    t_loc_ns=None,
    dt_range_ns=None,
    t_loc_default_ns=None,
) -> types.VectorOfVectors:
    """Calculate a 3D :class:`~lgdo.types.vectorofvectors.VectorOfVectors` pulse data mask.

    Useful to filter any pulse data based on pulse amplitude and start time.

    Parameters
    ----------
    datainfo, tcm, table_names
        positional arguments automatically supplied by :func:`.build_evt`.
    a_thr_pe
        amplitude threshold (in photoelectrons) used to build a pulse mask with
        :func:`make_pulse_data_mask`, if `pulse_mask` is ``None``. The output
        pulse data will be such that the pulse amplitude is above this value.
    t_loc_ns
        location of the time window in which pulses must sit. If a 1D array is
        provided, it is interpreted as a list of locations for each event (can
        be employed to e.g. provide the actual HPGe pulse position)
    dt_range_ns
        tuple with dimension of the time window in which pulses must sit
        relative to `t_loc_ns`. If, for example, `t_loc_ns` is 48000 ns and
        `dt_range_ns` is (-1000, 5000) ns, the resulting window will be (47000,
        53000) ns.
    t_loc_default_ns
        default value for `t_loc_ns`, in case the supplied value is
        :any:`numpy.nan`.
    """
    # get the t0 of each single pulse
    pulse_t0 = gather_pulse_data(
        datainfo,
        tcm,
        table_names,
        channel_mapping,
        observable="hit.trigger_pos",
        drop_empty=False,
    )

    pulse_amp = gather_pulse_data(
        datainfo,
        tcm,
        table_names,
        channel_mapping,
        observable="hit.energy_in_pe",
        drop_empty=False,
    ).view_as("ak")

    # (HPGe) trigger position can vary among events!
    if isinstance(t_loc_ns, types.Array):
        t_loc_ns = t_loc_ns.view_as("ak")

    if isinstance(t_loc_ns, ak.Array):
        if t_loc_ns.ndim != 1:
            msg = "t_loc_ns must be 0- or 1-dimensional"
            raise ValueError(msg)

        # NOTE: the assumption is that t0 is np.nan when missing -> replace
        # with default value
        t_loc_ns = ak.fill_none(ak.nan_to_none(t_loc_ns), t_loc_default_ns)

    # start with all-true mask
    pulse_t0_ns = pulse_t0.view_as("ak")
    mask = pulse_t0_ns == pulse_t0_ns

    # apply p.e. threshold
    if a_thr_pe is not None:
        mask = mask & (pulse_amp > a_thr_pe)

    # apply time windowing
    if t_loc_ns is not None and dt_range_ns is not None:
        if not isinstance(dt_range_ns, (tuple, list)):
            msg = "dt_range_ns must be a tuple"
            raise ValueError(msg)

        mask = mask & (
            (pulse_t0_ns < (t_loc_ns + dt_range_ns[1]))
            & (pulse_t0_ns > (t_loc_ns + dt_range_ns[0]))
        )

    return types.VectorOfVectors(mask)


def geds_coincidence_classifier(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    table_names: Sequence[str],
    channel_mapping: dict,
    *,
    spms_t0: types.VectorOfVectors,
    spms_amp: types.VectorOfVectors,
    geds_t0_ns: types.Array,
    ts_bkg_prob: float,
    rc_density: Sequence[float] | None = None,
) -> types.Array:
    """Calculate the HPGe / SiPMs coincidence classifier.

    The value represents the likelihood of a physical correlation between HPGe
    and SiPM signals.

    Parameters
    ----------
    datainfo, tcm, table_names
        positional arguments automatically supplied by :func:`.build_evt`.
    t0
        arrival times of pulses in ns, split by channel.
    amp
        amplitude of pulses in p.e., split by channel.
    geds_t0_ns
        t0 (ns) of the HPGe signal.
    ts_bkg_prob
        probability for a pulse coming from some uncorrelated physics (uniform
        distribution). needed for the LAr scintillation time pdf.
    rc_density
        density array of the random coincidence LAr energy distribution (total
        energy summed over all channels, in p.e.). Derived from forced trigger
        data.
    """
    return types.Array(
        larveto.l200_combined_test_stat(
            spms_t0.view_as("ak"),
            spms_amp.view_as("ak"),
            geds_t0_ns.view_as("ak"),
            ts_bkg_prob,
            rc_density,
        )
    )



"""Event processors for HPGe data."""

from __future__ import annotations

from collections.abc import Sequence

import awkward as ak
import numpy as np
from lgdo import lh5, types

from .. import utils
from . import xtalk


def apply_recovery_cut(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    table_names: Sequence[str],
    channel_mapping: dict,
    *,
    timestamps: types.Array,
    flag: types.Array,
    time_window: float,
) -> types.Array:

    discharge_timestamps = timestamps.nda[flag.nda == 1]
    is_recovering = np.full(len(timestamps.nda), False)
    for tstamp in discharge_timestamps:
        is_recovering = is_recovering | np.where(
            (
                ((timestamps.nda - tstamp) < time_window)
                & ((timestamps.nda - tstamp) >= 0)
            ),
            True,
            False,
        )

    # return the result as LGDO
    return types.Array(is_recovering)


def apply_xtalk_correction(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    table_names: Sequence[str],
    channel_mapping: dict,
    *,
    return_mode: str,
    uncal_energy_expr: str,
    cal_energy_expr: str,
    multiplicity_expr: str,
    xtalk_threshold: float = None,
    xtalk_matrix_filename: str = "",
    xtalk_rawid_obj: str = "xtc/rawid_index",
    xtalk_matrix_obj: str = "xtc/xtalk_matrix_negative",
    positive_xtalk_matrix_obj: str = "xtc/xtalk_matrix_positive",
) -> types.VectorOfVectors:
    """Applies the cross-talk correction to the energy observable.
    The format of `xtalk_matrix_filename` should be currently be a path to a lh5 file.

    The correction is applied using matrix algebra for all triggers above the threshold.

    Parameters
    ----------
    datainfo, tcm, table_names
        positional arguments automatically supplied by :func:`.build_evt`.
    return_mode
        string which can be either energy to return corrected energy or tcm_index
    uncal_energy_expr
        expression for the pulse parameter to be gathered for the uncalibrated energy (used for correction),
        can be a combination of different fields.
    cal_energy_expr
        expression for the pulse parameter to be gathered for the calibrated energy, used for the xtalk threshold,
        can be a combination of different fields.
    xtalk_threshold
        threshold used for xtalk correction, hits below this energy will not
        be used to correct the other hits.
    xtalk_matrix_filename
        name of the file containing the xtalk matrices.
    xtalk_matrix_obj
        name of the lh5 object containing the xtalk matrix
    positive_xtalk_matrix_obj
        name of the lh5 object containing the positive polarity xtalk matrix
    xtalk_rawids_obj
        name of the lh5 object containing the name of the rawids
    """

    xtalk_matrix_rawids = lh5.read_as(xtalk_rawid_obj, xtalk_matrix_filename, "np")
    tcm_index_array = xtalk.build_tcm_index_array(tcm, datainfo, xtalk_matrix_rawids)

    energy_corr = xtalk.get_xtalk_correction(
        tcm,
        datainfo,
        uncal_energy_expr,
        cal_energy_expr,
        xtalk_threshold,
        xtalk_matrix_filename,
        xtalk_rawid_obj,
        xtalk_matrix_obj,
        positive_xtalk_matrix_obj,
    )

    multiplicity_mask = xtalk.filter_hits(
        datainfo,
        tcm,
        multiplicity_expr,
        energy_corr,
        xtalk_matrix_rawids,
    )
    energy_corr = ak.from_regular(energy_corr)
    multiplicity_mask = ak.from_regular(multiplicity_mask)
    tcm_index_array = ak.from_regular(tcm_index_array)

    if return_mode == "energy":
        return types.VectorOfVectors(energy_corr[multiplicity_mask])
    elif return_mode == "tcm_index":
        return types.VectorOfVectors(tcm_index_array[multiplicity_mask])
    else:
        raise ValueError(f"Unknown mode: {return_mode}")


def apply_xtalk_correction_and_calibrate(
    datainfo: utils.DataInfo,
    tcm: utils.TCMData,
    table_names: Sequence[str],
    channel_mapping: dict,
    *,
    return_mode: str,
    uncal_energy_expr: str,
    cal_energy_expr: str,
    cal_par_files: str | Sequence[str],
    multiplicity_expr: str,
    xtalk_matrix_filename: str,
    xtalk_threshold: float = None,
    xtalk_rawid_obj: str = "xtc/rawid_index",
    xtalk_matrix_obj: str = "xtc/xtalk_matrix_negative",
    positive_xtalk_matrix_obj: str = "xtc/xtalk_matrix_positive",
    uncal_var: str = "dsp.cuspEmax",
    recal_var: str = "hit.cuspEmax_ctc_cal",
) -> types.VectorOfVectors:
    """Applies the cross-talk correction to the energy observable.

    The correction is applied using matrix algebra for all triggers above the
    xalk threshold.

    Parameters
    ----------
    datainfo, tcm, table_names
        positional arguments automatically supplied by :func:`.build_evt`.
    return_mode
        string which can be either ``energy`` to return corrected energy or
        ``tcm_index``.
    uncal_energy_expr
        expression for the pulse parameter to be gathered for the uncalibrated
        energy (used for correction), can be a combination of different fields.
    cal_energy_expr
        expression for the pulse parameter to be gathered for the calibrated
        energy, used for the xtalk threshold, can be a combination of different
        fields.
    cal_par_files
        path to the generated hit tier par-files defining the calibration
        curves. Used to recalibrate the data after xtalk correction.
    multiplicity_expr:
        expression defining the logic used to compute the event multiplicity.
    xtalk_threshold
        threshold used for xtalk correction, hits below this energy will not be
        used to correct the other hits.
    xtalk_matrix_filename
        path to the file containing the xtalk matrices.
    xtalk_matrix_obj
        name of the lh5 object containing the xtalk matrix.
    positive_xtalk_matrix_obj
        name of the lh5 object containing the positive polarity xtalk matrix.
    xtalk_matrix_rawids
        name of the lh5 object containing the name of the rawids.
    recal_var
        name of the energy variable to use for re-calibration.
    """

    xtalk_matrix_rawids = lh5.read_as(xtalk_rawid_obj, xtalk_matrix_filename, "np")
    tcm_index_array = xtalk.build_tcm_index_array(tcm, datainfo, xtalk_matrix_rawids)

    energy_corr = xtalk.get_xtalk_correction(
        tcm,
        datainfo,
        uncal_energy_expr,
        cal_energy_expr,
        xtalk_threshold,
        xtalk_matrix_filename,
        xtalk_rawid_obj,
        xtalk_matrix_obj,
        positive_xtalk_matrix_obj,
    )

    calibrated_corr = xtalk.calibrate_energy(
        datainfo,
        tcm,
        energy_corr,
        xtalk_matrix_rawids,
        cal_par_files,
        uncal_var,
        recal_var,
        channel_mapping,
    )

    multiplicity_mask = xtalk.filter_hits(
        datainfo,
        tcm,
        multiplicity_expr,
        calibrated_corr,
        xtalk_matrix_rawids,
    )

    calibrated_corr = ak.from_regular(calibrated_corr)
    multiplicity_mask = ak.from_regular(multiplicity_mask)
    tcm_index_array = ak.from_regular(tcm_index_array)

    if return_mode == "energy":
        return types.VectorOfVectors(calibrated_corr[multiplicity_mask])
    elif return_mode == "tcm_index":
        return types.VectorOfVectors(tcm_index_array[multiplicity_mask])
    else:
        raise ValueError(f"Unknown mode: {return_mode}")



import numba as nb
import numpy as np

from pygama.utils import numba_math_defaults as nb_defaults


@nb.njit(**nb_defaults(parallel=False))
def nb_poly(x: np.ndarray, pars: np.ndarray) -> np.ndarray:
    r"""
    A polynomial function with pars following the numpy polynomial convention. It computes:


    .. math::
        y = a_n x^n + ... + a_1 x + a_0


    As a Numba JIT function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        Input data
    pars
        Coefficients of the polynomial, in  numpy polynomial convention

    Returns
    -------
    result
        The polynomial defined by pars at the evaluated at x

    Notes
    -----
    This follows the :func:`numpy.polynomial` convention of :math:`a_0 + a_1 x +.... + a_n x^n`
    """

    result = x * 0  # do x*0 to keep shape of x (scalar or array)
    if len(pars) == 0:
        return result
    result += pars[0]
    y = x
    for i in nb.prange(1, len(pars)):
        result += pars[i] * x
        x = x * y
    return result



"""
This module provides functions for calculating survival fractions for a cut.
"""

from __future__ import annotations

import copy
import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from iminuit import Minuit, cost
from iminuit.util import ValueView

import pygama.pargen.energy_cal as pgc
from pygama.math.distributions import gauss_on_step, hpge_peak

log = logging.getLogger(__name__)


def energy_guess(energy, func_i, fit_range=None, bin_width=1, peak=None, eres=None):
    """
    Simple guess for peak fitting
    """
    if fit_range is None:
        fit_range = (np.nanmin(energy), np.nanmax(energy))
    if func_i == hpge_peak or func_i == gauss_on_step:
        parguess = pgc.get_hpge_energy_peak_par_guess(
            energy, func_i, fit_range=fit_range
        )

        if peak is not None:
            parguess["mu"] = peak

        if eres is not None:
            parguess["sigma"] = eres / 2.355

        for i, guess in enumerate(parguess):
            if np.isnan(guess):
                parguess[i] = 0

    else:
        log.error(f"energy_guess not implemented for {func_i}")
        return None
    return parguess


def fix_all_but_nevents(func):
    """
    Fixes all parameters except the number of signal and background events
    and their efficiencies
    Returns: Sequence list of fixed indexes for fitting and mask for parameters
    """

    if func == gauss_on_step:
        # pars are: n_sig, mu, sigma, n_bkg, hstep, lower, upper, components
        fixed = ["x_lo", "x_hi", "mu", "sigma", "hstep"]

    elif func == hpge_peak:
        # pars are: , components
        fixed = ["x_lo", "x_hi", "mu", "sigma", "htail", "tau", "hstep"]

    else:
        log.error(f"get_hpge_E_fixed not implemented for {func}")
        return None, None
    mask = ~np.in1d(func.required_args(), fixed)
    return fixed, mask


def get_bounds(func, parguess):
    """
    Gets bounds for the fit parameters
    """
    if func == hpge_peak or func == gauss_on_step:
        bounds = pgc.get_hpge_energy_bounds(func, parguess)

        bounds["mu"] = (parguess["mu"] - 1, parguess["mu"] + 1)
        bounds["n_sig"] = (0, 2 * (parguess["n_sig"] + parguess["n_bkg"]))
        bounds["n_bkg"] = (0, 2 * (parguess["n_sig"] + parguess["n_bkg"]))

    else:
        log.error(f"get_bounds not implemented for {func}")
        return None
    return bounds


def pass_pdf_gos(
    x, x_lo, x_hi, n_sig, epsilon_sig, n_bkg, epsilon_bkg, mu, sigma, hstep1, hstep2
):
    """
    pdf for gauss on step fit reparamerised to calculate the efficiency of the cut
    this is the passing pdf
    """
    return gauss_on_step.pdf_ext(
        x, x_lo, x_hi, n_sig * epsilon_sig, mu, sigma, n_bkg * epsilon_bkg, hstep1
    )


def fail_pdf_gos(
    x, x_lo, x_hi, n_sig, epsilon_sig, n_bkg, epsilon_bkg, mu, sigma, hstep1, hstep2
):
    """
    pdf for gauss on step fit reparamerised to calculate the efficiency of the cut
    this is the cut pdf
    """
    return gauss_on_step.pdf_ext(
        x,
        x_lo,
        x_hi,
        n_sig * (1 - epsilon_sig),
        mu,
        sigma,
        n_bkg * (1 - epsilon_bkg),
        hstep2,
    )


def pass_pdf_hpge(
    x,
    x_lo,
    x_hi,
    n_sig,
    epsilon_sig,
    n_bkg,
    epsilon_bkg,
    mu,
    sigma,
    htail,
    tau,
    hstep1,
    hstep2,
):
    """
    pdf for hpge peak fit reparamerised to calculate the efficiency of the cut.
    this is the passing pdf
    """
    return hpge_peak.pdf_ext(
        x,
        x_lo,
        x_hi,
        n_sig * epsilon_sig,
        mu,
        sigma,
        htail,
        tau,
        n_bkg * epsilon_bkg,
        hstep1,
    )


def fail_pdf_hpge(
    x,
    x_lo,
    x_hi,
    n_sig,
    epsilon_sig,
    n_bkg,
    epsilon_bkg,
    mu,
    sigma,
    htail,
    tau,
    hstep1,
    hstep2,
):
    """
    pdf for hpge peak fit reparamerised to calculate the efficiency of the cut
    this is the cut pdf
    """
    return hpge_peak.pdf_ext(
        x,
        x_lo,
        x_hi,
        n_sig * (1 - epsilon_sig),
        mu,
        sigma,
        htail,
        tau,
        n_bkg * (1 - epsilon_bkg),
        hstep2,
    )


def update_guess(func, parguess, energies):
    """
    Updates guess for the number of signal and background events
    """
    if func == gauss_on_step or func == hpge_peak:

        total_events = len(energies)
        parguess["n_sig"] = len(
            energies[
                (energies > parguess["mu"] - 2 * parguess["sigma"])
                & (energies < parguess["mu"] + 2 * parguess["sigma"])
            ]
        )
        parguess["n_sig"] -= len(
            energies[
                (energies > parguess["x_lo"])
                & (energies < parguess["x_lo"] + 2 * parguess["sigma"])
            ]
        )
        parguess["n_sig"] -= len(
            energies[
                (energies > parguess["x_hi"] - 2 * parguess["sigma"])
                & (energies < parguess["x_hi"])
            ]
        )
        parguess["n_bkg"] = total_events - parguess["n_sig"]
        return parguess

    else:
        log.error(f"update_guess not implemented for {func}")
        return parguess


def get_survival_fraction(
    energy: np.ndarray,
    cut_param: np.ndarray,
    cut_val: float,
    peak: float,
    eres_pars: float,
    fit_range: tuple = None,
    high_cut: float = None,
    pars: ValueView = None,
    data_mask: np.ndarray = None,
    mode: str = "greater",
    func=hpge_peak,
    fix_step=True,
    display=0,
):
    """
    Function for calculating the survival fraction of a cut
    using a fit to the surviving and failing energy distributions.

    Parameters
    ----------

    energy: array
        array of energies
    cut_param: array
        array of the cut parameter for the survival fraction calculation, should have the same length as energy
    cut_val: float
        value of the cut parameter to be used for the survival fraction calculation
    peak: float
        energy of the peak to be fitted
    eres_pars: float
        energy resolution parameter for the peak
    fit_range: tuple
        range of the fit in keV
    high_cut: float
        upper value for the cut parameter to have a range in the cut value
    pars: iMinuit ValueView
        initial parameters for the fit
    data_mask: array
        mask for the data to be used in the fit
    mode: str
        mode of the cut, either "greater" or "less"
    func: function
        function to be used in the fit
    fix_step: bool
        option to fix the step parameters in the fit
    display: int
        option to display the fit if greater than 1

    Returns
    -------

    sf: float
        survival fraction
    err: float
        error on the survival fraction
    values: iMinuit ValueView
        values of the parameters of the fit
    errors: iMinuit ValueView
        errors on the parameters of the fit
    """
    if data_mask is None:
        data_mask = np.full(len(cut_param), True, dtype=bool)

    if not isinstance(energy, np.ndarray):
        energy = np.array(energy)
    if not isinstance(cut_param, np.ndarray):
        cut_param = np.array(cut_param)

    if fit_range is None:
        fit_range = (np.nanmin(energy), np.nanmax(energy))

    nan_idxs = np.isnan(cut_param)
    if high_cut is not None:
        idxs = (cut_param > cut_val) & (cut_param < high_cut) & data_mask
    else:
        if mode == "greater":
            idxs = (cut_param > cut_val) & data_mask
        elif mode == "less":
            idxs = (cut_param < cut_val) & data_mask
        else:
            raise ValueError("mode not recognised")

    if pars is None:
        (pars, errs, cov, _, func, _, _, _) = pgc.unbinned_staged_energy_fit(
            energy,
            func,
            guess_func=energy_guess,
            bounds_func=get_bounds,
            guess_kwargs={"peak": peak, "eres": eres_pars},
            fit_range=fit_range,
        )

    guess_pars_surv = copy.deepcopy(pars)

    # add update guess here for n_sig and n_bkg
    guess_pars_surv = update_guess(func, guess_pars_surv, energy[(~nan_idxs) & (idxs)])

    parguess = {
        "x_lo": pars["x_lo"],
        "x_hi": pars["x_hi"],
        "mu": pars["mu"],
        "sigma": pars["sigma"],
        "hstep1": pars["hstep"],
        "hstep2": pars["hstep"],
        "n_sig": pars["n_sig"],
        "n_bkg": pars["n_bkg"],
        "epsilon_sig": guess_pars_surv["n_sig"] / pars["n_sig"],
        "epsilon_bkg": guess_pars_surv["n_bkg"] / pars["n_bkg"],
    }

    bounds = {
        "n_sig": (0, pars["n_sig"] + pars["n_bkg"]),
        "epsilon_sig": (0, 1),
        "n_bkg": (0, pars["n_bkg"] + pars["n_sig"]),
        "epsilon_bkg": (0, 1),
        "hstep1": (-1, 1),
        "hstep2": (-1, 1),
    }

    if func == hpge_peak:
        parguess.update({"htail": pars["htail"], "tau": pars["tau"]})

    if func == hpge_peak:
        lh = cost.ExtendedUnbinnedNLL(
            energy[(~nan_idxs) & (idxs)], pass_pdf_hpge
        ) + cost.ExtendedUnbinnedNLL(energy[(~nan_idxs) & (~idxs)], fail_pdf_hpge)
    elif func == gauss_on_step:
        lh = cost.ExtendedUnbinnedNLL(
            energy[(~nan_idxs) & (idxs)], pass_pdf_gos
        ) + cost.ExtendedUnbinnedNLL(energy[(~nan_idxs) & (~idxs)], fail_pdf_gos)

    else:
        raise ValueError("Unknown func")

    m = Minuit(lh, **parguess)
    fixed = ["x_lo", "x_hi", "n_sig", "n_bkg", "mu", "sigma"]  # "hstep"
    if func == hpge_peak:
        fixed += ["tau", "htail"]
    if fix_step is True:
        fixed += ["hstep1", "hstep2"]

    m.fixed[fixed] = True
    for arg, val in bounds.items():
        m.limits[arg] = val

    m.simplex().migrad()
    m.hesse()

    sf = m.values["epsilon_sig"] * 100
    err = m.errors["epsilon_sig"] * 100

    if display > 1:
        fig, (ax1, ax2) = plt.subplots(1, 2)
        bins = np.arange(1552, 1612, 1)
        ax1.hist(energy[(~nan_idxs) & (idxs)], bins=bins, histtype="step")

        ax2.hist(energy[(~nan_idxs) & (~idxs)], bins=bins, histtype="step")

        if func == hpge_peak:
            ax1.plot(bins, pass_pdf_hpge(bins, **m.values.to_dict())[1])
            ax2.plot(bins, fail_pdf_hpge(bins, **m.values.to_dict())[1])
        elif func == gauss_on_step:
            ax1.plot(bins, pass_pdf_gos(bins, **m.values.to_dict())[1])
            ax2.plot(bins, fail_pdf_gos(bins, **m.values.to_dict())[1])

        plt.show()

    return sf, err, m.values, m.errors


def get_sf_sweep(
    energy: np.array,
    cut_param: np.array,
    final_cut_value: float = None,
    peak: float = 1592.5,
    eres_pars: list = None,
    data_mask=None,
    cut_range=(-5, 5),
    n_samples=26,
    mode="greater",
    fit_range=None,
    debug_mode=False,
) -> tuple(pd.DataFrame, float, float):
    """
    Function sweeping through cut values and calculating the survival fraction for each value
    using a fit to the surviving and failing enegry distributions.

    Parameters
    ----------

    energy: array
        array of energies
    cut_param: array
        array of the cut parameter for the survival fraction calculation, should have the same length as energy
    final_cut_val: float
        value of the final cut parameter to be used for the survival fraction calculation
    peak: float
        energy of the peak to be fitted
    eres_pars: float
        energy resolution parameter for the peak
    data_mask: array
        mask for the data to be used in the fit
    cut_range: tuple
        range of the cut values to be swept through
    n_samples: int
        number of samples to be taken in the cut range
    mode: str
        mode of the cut, either "greater" or "less"
    fit_range: tuple
        range of the fit in keV
    debug_mode: bool
        option to raise an error if there is an issue with

    Returns
    -------

    out_df: Dataframe
        Dataframe of cut values with the survival fraction and error
    sf: float
        survival fraction
    err: float
        error on the survival fraction

    """

    if data_mask is None:
        data_mask = np.full(len(cut_param), True, dtype=bool)

    if not isinstance(energy, np.ndarray):
        energy = np.array(energy)
    if not isinstance(cut_param, np.ndarray):
        cut_param = np.array(cut_param)

    cut_vals = np.linspace(cut_range[0], cut_range[1], n_samples)
    out_df = pd.DataFrame()

    (pars, errs, _, _, func, _, _, _) = pgc.unbinned_staged_energy_fit(
        energy,
        hpge_peak,
        guess_func=energy_guess,
        bounds_func=get_bounds,
        guess_kwargs={"peak": peak, "eres": eres_pars},
        fit_range=fit_range,
    )

    for cut_val in cut_vals:
        try:
            sf, err, _, _ = get_survival_fraction(
                energy,
                cut_param,
                cut_val,
                peak,
                eres_pars,
                fit_range=fit_range,
                data_mask=data_mask,
                mode=mode,
                pars=pars,
                func=func,
            )
            out_df = pd.concat(
                [out_df, pd.DataFrame([{"cut_val": cut_val, "sf": sf, "sf_err": err}])]
            )
        except BaseException as e:
            if e == KeyboardInterrupt:
                raise (e)
            elif debug_mode:
                raise (e)
    out_df.set_index("cut_val", inplace=True)
    if final_cut_value is not None:
        sf, sf_err, _, _ = get_survival_fraction(
            energy,
            cut_param,
            final_cut_value,
            peak,
            eres_pars,
            fit_range=fit_range,
            data_mask=data_mask,
            mode=mode,
            pars=pars,
            func=func,
        )
    else:
        sf = None
        sf_err = None
    return (
        out_df,
        sf,
        sf_err,
    )


def compton_sf(
    cut_param, low_cut_val, high_cut_val=None, mode="greater", data_mask=None
) -> dict:
    """
    Function for calculating the survival fraction of a cut in a compton region with
    a simple counting analysis.

    Parameters
    ----------

    cut_param: array
        array of the cut parameter for the survival fraction calculation, should have the same length as energy
    low_cut_val: float
        value of the cut parameter to be used for the survival fraction calculation
    high_cut_val: float
        upper value for the cut parameter to have a range in the cut value
    mode: str
        mode of the cut, either "greater" or "less"
    data_mask: array
        mask for the data to be used in the fit

    Returns
    -------

    sf : dict
        dictionary containing the low cut value, survival fraction and error on the survival fraction
    """
    if data_mask is None:
        data_mask = np.full(len(cut_param), True, dtype=bool)

    if not isinstance(cut_param, np.ndarray):
        cut_param = np.array(cut_param)

    if high_cut_val is not None:
        mask = (cut_param > low_cut_val) & (cut_param < high_cut_val) & data_mask
    else:
        if mode == "greater":
            mask = (cut_param > low_cut_val) & data_mask
        elif mode == "less":
            mask = (cut_param < low_cut_val) & data_mask
        else:
            raise ValueError("mode not recognised")

    sf = len(cut_param[mask]) / len(cut_param)
    err = 100 * np.sqrt((sf * (1 - sf)) / len(cut_param))
    sf *= 100

    return {
        "low_cut": low_cut_val,
        "sf": sf,
        "sf_err": err,
        "high_cut": high_cut_val,
    }


def compton_sf_sweep(
    energy: np.array,
    cut_param: np.array,
    final_cut_value: float,
    data_mask: np.array = None,
    cut_range=(-5, 5),
    n_samples=51,
    mode="greater",
) -> tuple(pd.DataFrame, float, float):
    """
    Function sweeping through cut values and calculating the survival fraction for each value
    using a simple counting analysis.

    Parameters
    ----------

    energy: array
        array of energies
    cut_param: array
        array of the cut parameter for the survival fraction calculation, should have the same length as energy
    final_cut_val: float
        value of the final cut parameter to be used for the survival fraction calculation
    data_mask: array
        mask for the data to be used in the fit
    cut_range: tuple
        range of the cut values to be swept through
    n_samples: int
        number of samples to be taken in the cut range
    mode: str
        mode of the cut, either "greater" or "less"

    Returns
    -------

    out_df: Dataframe
        Dataframe of cut values with the survival fraction and error
    sf: float
        survival fraction
    err: float
        error on the survival fraction

    """
    if not isinstance(energy, np.ndarray):
        energy = np.array(energy)
    if not isinstance(cut_param, np.ndarray):
        cut_param = np.array(cut_param)

    cut_vals = np.linspace(cut_range[0], cut_range[1], n_samples)
    out_df = pd.DataFrame()

    for cut_val in cut_vals:
        ct_dict = compton_sf(cut_param, cut_val, mode=mode, data_mask=data_mask)
        df = pd.DataFrame(
            [
                {
                    "cut_val": ct_dict["low_cut"],
                    "sf": ct_dict["sf"],
                    "sf_err": ct_dict["sf_err"],
                }
            ]
        )
        out_df = pd.concat([out_df, df])
    out_df.set_index("cut_val", inplace=True)

    sf_dict = compton_sf(cut_param, final_cut_value, mode=mode, data_mask=data_mask)

    return out_df, sf_dict["sf"], sf_dict["sf_err"]



import numpy as np
from numpy.polynomial.polynomial import Polynomial

from pygama.math.functions.polynomial import nb_poly


def test_nb_poly():
    x = np.arange(-10, 10)
    params = np.array([1, 2, 3])
    y = nb_poly(x, params)
    np_poly = Polynomial(params)
    y_numpy = np_poly(x)

    assert np.array_equal(y, y_numpy)



import os

import lgdo
import numpy as np
import pytest
from lgdo import lh5
from lgdo.types import Table, VectorOfVectors

from pygama import evt


def test_generate_tcm_cols(lgnd_test_data):
    f_raw = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_raw.lh5"
    )

    tcm_cols = evt.build_tcm(
        [(f_raw, f"{chan}/raw") for chan in lh5.ls(f_raw)],
        "timestamp",
        buffer_len=100,
    )

    assert isinstance(tcm_cols, Table)
    assert isinstance(tcm_cols.table_key, VectorOfVectors)
    assert isinstance(tcm_cols.row_in_table, VectorOfVectors)
    for v in tcm_cols.values():
        assert np.issubdtype(v.flattened_data.nda.dtype, np.integer)

    # test attrs
    assert set(tcm_cols.attrs.keys()) == {"datatype", "hash_func", "tables"}
    assert tcm_cols.attrs["hash_func"] == r"\d+"
    assert set(eval(tcm_cols.attrs["tables"])) == {
        f"{chan}/raw" for chan in lh5.ls(f_raw)
    }

    # fmt: off
    assert np.array_equal(
        tcm_cols.table_key.cumulative_length.nda,
        [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
        ],
    )
    assert np.array_equal(
        tcm_cols.table_key.flattened_data.nda,
        [
            1084804, 1084803, 1121600, 1084804, 1121600, 1084804, 1121600,
            1084804, 1084804, 1084804, 1084803, 1084804, 1084804, 1121600,
            1121600, 1084804, 1121600, 1084804, 1121600, 1084803, 1084803,
            1121600, 1121600, 1121600, 1084803, 1084803, 1084803, 1084803,
            1084803, 1084803,
        ],
    )
    assert np.array_equal(
        tcm_cols.row_in_table.flattened_data.nda,
        [
            0, 0, 0, 1, 1, 2, 2, 3, 4, 5, 1, 6, 7, 3, 4, 8, 5, 9, 6, 2, 3, 7,
            8, 9, 4, 5, 6, 7, 8, 9,
        ],
    )

    # fmt: on
    # test with small buffer len
    tcm_cols = evt.build_tcm(
        [(f_raw, f"{chan}/raw") for chan in lh5.ls(f_raw)],
        "timestamp",
        buffer_len=1,
    )

    assert isinstance(tcm_cols, Table)
    assert isinstance(tcm_cols.table_key, VectorOfVectors)
    assert isinstance(tcm_cols.row_in_table, VectorOfVectors)
    for v in tcm_cols.values():
        assert np.issubdtype(v.flattened_data.nda.dtype, np.integer)
    # fmt: off
    assert np.array_equal(
        tcm_cols.table_key.cumulative_length.nda,
        [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
        ],
    )
    assert np.array_equal(
        tcm_cols.table_key.flattened_data.nda,
        [
            1084804, 1084803, 1121600, 1084804, 1121600, 1084804, 1121600,
            1084804, 1084804, 1084804, 1084803, 1084804, 1084804, 1121600,
            1121600, 1084804, 1121600, 1084804, 1121600, 1084803, 1084803,
            1121600, 1121600, 1121600, 1084803, 1084803, 1084803, 1084803,
            1084803, 1084803,
        ],
    )
    assert np.array_equal(
        tcm_cols.row_in_table.flattened_data.nda,
        [
            0, 0, 0, 1, 1, 2, 2, 3, 4, 5, 1, 6, 7, 3, 4, 8, 5, 9, 6, 2, 3, 7,
            8, 9, 4, 5, 6, 7, 8, 9,
        ],
    )
    # fmt: on
    # test with None hash_func
    tcm_cols = evt.build_tcm(
        [(f_raw, f"{chan}/raw") for chan in lh5.ls(f_raw)],
        "timestamp",
        hash_func=None,
        buffer_len=1,
    )
    # fmt: off
    assert np.array_equal(
        tcm_cols.table_key.flattened_data.nda,
        [
            1, 0, 2, 1, 2, 1, 2, 1, 1, 1, 0,
            1, 1, 2, 2, 1, 2, 1, 2, 0, 0, 2,
            2, 2, 0, 0, 0, 0, 0, 0,
        ],
    )
    # fmt: on
    # test invalid hash func
    with pytest.raises(NotImplementedError):
        evt.build_tcm(
            [(f_raw, f"{chan}/raw") for chan in lh5.ls(f_raw)],
            "timestamp",
            hash_func=[],
        )

    # test invalid window_refs
    with pytest.raises(NotImplementedError):
        evt.build_tcm(
            [(f_raw, f"{chan}/raw") for chan in lh5.ls(f_raw)],
            "timestamp",
            window_refs="test",
        )

    # test adding extra fields
    tcm_cols = evt.build_tcm(
        [(f_raw, f"{chan}/raw") for chan in lh5.ls(f_raw)],
        "timestamp",
        out_fields="timestamp",
    )
    assert "timestamp" in tcm_cols.keys()

    # test channel appearing multiple times in single entry
    tcm_cols = evt.build_tcm(
        [(f_raw, f"{chan}/raw") for chan in lh5.ls(f_raw)],
        "timestamp",
        buffer_len=100,
        coin_windows=1,
    )
    assert np.array_equal(
        tcm_cols.table_key.cumulative_length.nda,
        [30],
    )
    # fmt: off
    assert np.array_equal(
        tcm_cols.table_key.flattened_data.nda,
        [
            1084804, 1084803, 1121600, 1084804, 1121600, 1084804,
            1121600, 1084804, 1084804, 1084804, 1084803, 1084804,
            1084804, 1121600, 1121600, 1084804, 1121600, 1084804,
            1121600, 1084803, 1084803, 1121600, 1121600, 1121600,
            1084803, 1084803, 1084803, 1084803, 1084803, 1084803,
        ],
    )
    # fmt: on


def test_build_tcm_multiple_cols(lgnd_test_data):
    f_raw = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_raw.lh5"
    )

    with pytest.raises(ValueError):
        evt.build_tcm(
            [(f_raw, ["ch1084803/raw", "ch1084804/raw", "ch1121600/raw"])],
            coin_cols="timestamp",
            window_refs=["last", "last"],
        )
    with pytest.raises(ValueError):
        evt.build_tcm(
            [(f_raw, ["ch1084803/raw", "ch1084804/raw", "ch1121600/raw"])],
            coin_cols=["timestamp"],
            coin_windows=[1, 2],
        )
    tcm = evt.build_tcm(
        [(f_raw, ["ch1084803/raw", "ch1084804/raw", "ch1121600/raw"])],
        coin_cols=["timestamp", "table_key"],
    )
    assert isinstance(tcm, Table)
    assert isinstance(tcm.table_key, VectorOfVectors)
    assert isinstance(tcm.row_in_table, VectorOfVectors)
    # fmt: off
    assert np.array_equal(
        tcm.table_key.flattened_data.nda,
        [
            1084804, 1084803, 1121600, 1084804, 1121600, 1084804, 1121600,
            1084804, 1084804, 1084804, 1084803, 1084804, 1084804, 1121600,
            1121600, 1084804, 1121600, 1084804, 1121600, 1084803, 1084803,
            1121600, 1121600, 1121600, 1084803, 1084803, 1084803, 1084803,
            1084803, 1084803,
        ],
    )
    # fmt: on

    assert np.array_equal(
        tcm.table_key.cumulative_length.nda,
        np.arange(1, 31),
    )


def test_build_tcm_write(lgnd_test_data, tmp_dir):
    f_raw = lgnd_test_data.get_path(
        "lh5/prod-ref-l200/generated/tier/raw/cal/p03/r001/l200-p03-r001-cal-20230318T012144Z-tier_raw.lh5"
    )
    out_file = f"{tmp_dir}/pygama-test-tcm.lh5"
    evt.build_tcm(
        [(f_raw, ["ch1084803/raw", "ch1084804/raw", "ch1121600/raw"])],
        "timestamp",
        out_file=out_file,
        out_name="hardware_tcm",
        wo_mode="of",
    )

    assert os.path.exists(out_file)
    tcm_cols = lh5.read("hardware_tcm", out_file)
    assert isinstance(tcm_cols, lgdo.Struct)
    assert sorted(tcm_cols.keys()) == ["row_in_table", "table_key"]
    # fmt: off
    assert np.array_equal(
        tcm_cols.table_key.cumulative_length.nda,
        [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
        ],
    )
    assert np.array_equal(
        tcm_cols.table_key.flattened_data.nda,
        [
            1084804, 1084803, 1121600, 1084804, 1121600, 1084804, 1121600,
            1084804, 1084804, 1084804, 1084803, 1084804, 1084804, 1121600,
            1121600, 1084804, 1121600, 1084804, 1121600, 1084803, 1084803,
            1121600, 1121600, 1121600, 1084803, 1084803, 1084803, 1084803,
            1084803, 1084803,
        ],
    )
    assert np.array_equal(
        tcm_cols.row_in_table.flattened_data.nda,
        [
            0, 0, 0, 1, 1, 2, 2, 3, 4, 5, 1, 6, 7, 3, 4, 8, 5, 9, 6, 2, 3, 7,
            8, 9, 4, 5, 6, 7, 8, 9,
        ],
    )
    # fmt: on
    # test also with small buffers
    evt.build_tcm(
        [(f_raw, ["ch1084803/raw", "ch1084804/raw", "ch1121600/raw"])],
        "timestamp",
        out_file=out_file,
        out_name="hardware_tcm",
        wo_mode="of",
        buffer_len=1,
    )
    assert os.path.exists(out_file)
    tcm_cols = lh5.read("hardware_tcm", out_file)
    assert isinstance(tcm_cols, lgdo.Struct)
    assert sorted(tcm_cols.keys()) == ["row_in_table", "table_key"]
    # fmt: off
    assert np.array_equal(
        tcm_cols.table_key.cumulative_length.nda,
        [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
        ],
    )
    assert np.array_equal(
        tcm_cols.table_key.flattened_data.nda,
        [
            1084804, 1084803, 1121600, 1084804, 1121600, 1084804, 1121600,
            1084804, 1084804, 1084804, 1084803, 1084804, 1084804, 1121600,
            1121600, 1084804, 1121600, 1084804, 1121600, 1084803, 1084803,
            1121600, 1121600, 1121600, 1084803, 1084803, 1084803, 1084803,
            1084803, 1084803,
        ],
    )
    assert np.array_equal(
        tcm_cols.row_in_table.flattened_data.nda,
        [
            0, 0, 0, 1, 1, 2, 2, 3, 4, 5, 1, 6, 7, 3, 4, 8, 5, 9, 6, 2, 3, 7,
            8, 9, 4, 5, 6, 7, 8, 9,
        ],
    )
    # fmt: on
    # test also with small buffers
    evt.build_tcm(
        [(f_raw, ["ch1084803/raw", "ch1084804/raw", "ch1121600/raw"])],
        "timestamp",
        out_file=out_file,
        out_name="hardware_tcm",
        wo_mode="of",
        buffer_len=1,
    )
    assert os.path.exists(out_file)
    tcm_cols = lh5.read("hardware_tcm", out_file)
    assert isinstance(tcm_cols, lgdo.Struct)
    assert sorted(tcm_cols.keys()) == ["row_in_table", "table_key"]
    # fmt: off
    assert np.array_equal(
        tcm_cols.table_key.cumulative_length.nda,
        [
            1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
        ],
    )
    assert np.array_equal(
        tcm_cols.table_key.flattened_data.nda,
        [
            1084804, 1084803, 1121600, 1084804, 1121600, 1084804, 1121600,
            1084804, 1084804, 1084804, 1084803, 1084804, 1084804, 1121600,
            1121600, 1084804, 1121600, 1084804, 1121600, 1084803, 1084803,
            1121600, 1121600, 1121600, 1084803, 1084803, 1084803, 1084803,
            1084803, 1084803,
        ],
    )
    assert np.array_equal(
        tcm_cols.row_in_table.flattened_data.nda,
        [
            0, 0, 0, 1, 1, 2, 2, 3, 4, 5, 1, 6, 7, 3, 4, 8, 5, 9, 6, 2, 3, 7,
            8, 9, 4, 5, 6, 7, 8, 9,
        ],
    )
    # fmt: on



"""
pygama utility functions.
"""

import logging
from typing import Callable, Optional

import numpy as np

log = logging.getLogger(__name__)


def sizeof_fmt(num: float, suffix: str = "B") -> str:
    """
    given a file size in bytes, output a human-readable form.
    Parameters
    ----------
    num
        File size, in bytes
    suffix
        Desired file size suffix
    """
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            return f"{num:.3f} {unit}{suffix}"
        num /= 1024.0
    return "{:.1f} {} {}".format(num, "Y", suffix)


def get_par_names(func: Callable) -> tuple[str, ...]:
    """
    Return a list containing the names of the arguments of "func" other than the
    first argument. In pygamaland, those are the function's "parameters."

    Parameters
    ----------
    func
        A function whose parameters we want to return
    """
    import inspect

    par = inspect.getfullargspec(func)
    return par[0][1:]


def get_formatted_stats(mean: float, sigma: float, ndigs: int = 2) -> str:
    """
    convenience function for formatting mean +/- sigma to the right number of
    significant figures.

    Parameters
    ----------
    mean
        The mean value we want to format
    sigma
        The sigma value we want to format
    ndigs
        The number of significant digits we want to display
    """
    if sigma == 0:
        fmt = "%d" % ndigs
        fmt = "%#." + fmt + "g"
        return fmt % mean, fmt % sigma
    sig_pos = int(np.floor(np.log10(abs(sigma))))
    sig_fmt = "%d" % ndigs
    sig_fmt = "%#." + sig_fmt + "g"
    mean_pos = int(np.floor(np.log10(abs(mean))))
    mdigs = mean_pos - sig_pos + ndigs
    if mdigs < ndigs - 1:
        mdigs = ndigs - 1
    mean_fmt = "%d" % mdigs
    mean_fmt = "%#." + mean_fmt + "g"
    return mean_fmt % mean, sig_fmt % sigma


def print_fit_results(
    pars: np.ndarray,
    cov: np.ndarray,
    func: Optional[Callable] = None,
    title: Optional[str] = None,
    pad: Optional[bool] = True,
) -> None:
    """
    Convenience function to write scipy.optimize.curve_fit results to the log

    Parameters
    ----------
    pars
        The parameter values of the function func
    func
        A function, if passed then the function's parameters' names are logged
    title
        A title to log
    pad
        If True, adds spaces to the log messages

    Returns
    -------
    None
        Writes the curve_fit results to the log
    """
    if title is not None:
        log.info(f"{title}:")
    par_names = []
    if func is None:
        for i in range(len(pars)):
            par_names.append("p" + str(i))
    else:
        par_names = get_par_names(func)
    for i in range(len(pars)):
        mean, sigma = get_formatted_stats(pars[i], np.sqrt(cov[i][i]))
        log.info(f"{par_names[i]} = {mean} +/- {sigma}")
    if pad:
        log.info("")



"""
Crystal ball distributions for Pygama
"""

from math import erf

import numba as nb
import numpy as np

from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.utils import numba_math_defaults as nb_defaults
from pygama.utils import numba_math_defaults_kwargs as nb_kwargs


@nb.njit(**nb_kwargs)
def nb_crystal_ball_pdf(
    x: np.ndarray, mu: float, sigma: float, beta: float, m: float
) -> np.ndarray:
    r"""
    PDF of a power-law tail plus gaussian. Its range of support is :math:`x\in\mathbb{R}, \beta>0, m>1`. It computes:


    .. math::
        pdf(x, \beta, m, \mu, \sigma) =  \begin{cases}NA(B-\frac{x-\mu}{\sigma})^{-m} \quad \frac{x-\mu}{\sigma}\leq -\beta \\ Ne^{-(\frac{x-\mu}{\sigma})^2/2} \quad \frac{x-\mu}{\sigma}>-\beta\end{cases}


    Where


    .. math::
        A =  \frac{m^m}{\beta^m}e^{-\beta^2/2} \\
        B = \frac{m}{\beta}-\beta \\
        n =  \frac{1}{\sigma \frac{m e^{-\beta^2/2}}{\beta(m-1)} + \sigma \sqrt{\frac{\pi}{2}}\left(1+\text{erf}\left(\frac{\beta}{\sqrt{2}}\right)\right)}


    As a Numba vectorized function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    mu
        The amount to shift the distribution
    sigma
        The amount to scale the distribution
    beta
        The point where the pdf changes from power-law to Gaussian
    m
        The power of the power-law tail
    """

    if (beta <= 0) or (m <= 1):
        raise ValueError("beta must be greater than 0, and m must be greater than 1")

    # Define some constants to calculate the function
    const_a = (m / np.abs(beta)) ** m * np.exp(-1 * beta**2 / 2.0)
    const_b = m / np.abs(beta) - np.abs(beta)

    n = 1.0 / (
        m / np.abs(beta) / (m - 1) * np.exp(-(beta**2) / 2.0)
        + np.sqrt(np.pi / 2) * (1 + erf(np.abs(beta) / np.sqrt(2.0)))
    )

    y = np.empty_like(x, dtype=np.float64)
    for i in nb.prange(x.shape[0]):
        # Shift the distribution
        y[i] = (x[i] - mu) / sigma
        # Check if it is powerlaw
        if y[i] <= -1 * beta:
            y[i] = n * const_a * (const_b - y[i]) ** (-1 * m) / sigma
        # If it isn't power law, then it Gaussian
        else:
            y[i] = n * np.exp(-1 * y[i] ** 2 / 2) / sigma

    return y


@nb.njit(**nb_kwargs)
def nb_crystal_ball_cdf(
    x: np.ndarray, mu: float, sigma: float, beta: float, m: float
) -> np.ndarray:
    r"""
    CDF for power-law tail plus gaussian. Its range of support is :math:`x\in\mathbb{R}, \beta>0, m>1`. It computes:


    .. math::
        cdf(x, \beta, m,  \mu, \sigma)= \begin{cases}  NA\sigma\frac{(B-\frac{x-\mu}{\sigma})^{1-m}}{m-1} \quad , \frac{x-\mu}{\sigma} \leq -\beta \\ NA\sigma\frac{(B+\beta)^{1-m}}{m-1} + n\sigma \sqrt{\frac{\pi}{2}}\left(\text{erf}\left(\frac{x-\mu}{\sigma \sqrt{2}}\right)+\text{erf}\left(\frac{\beta}{\sqrt{2}}\right)\right)  \quad , \frac{x-\mu}{\sigma} >  -\beta \end{cases}


    Where


    .. math::
        A =  \frac{m^m}{\beta^m}e^{-\beta^2/2} \\
        B = \frac{m}{\beta}-\beta \\
        n =  \frac{1}{\sigma \frac{m e^{-\beta^2/2}}{\beta(m-1)} + \sigma \sqrt{\frac{\pi}{2}}\left(1+\text{erf}\left(\frac{\beta}{\sqrt{2}}\right)\right)}


    As a Numba vectorized function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    mu
        The amount to shift the distribution
    sigma
        The amount to scale the distribution
    beta
        The point where the cdf changes from power-law to Gaussian
    m
        The power of the power-law tail
    """

    if (beta <= 0) or (m <= 1):
        raise ValueError("beta must be greater than 0, and m must be greater than 1")
    # Define some constants to calculate the function
    const_a = (m / np.abs(beta)) ** m * np.exp(-1 * beta**2 / 2.0)
    const_b = m / np.abs(beta) - np.abs(beta)

    # Calculate the normalization constant
    n = 1.0 / (
        (np.sqrt(np.pi / 2) * (erf(beta / np.sqrt(2)) + 1))
        + ((const_a * (const_b + beta) ** (1 - m)) / (m - 1))
    )

    y = np.empty_like(x, dtype=np.float64)

    # Check if it is in the power law part
    for i in nb.prange(x.shape[0]):
        # Shift the distribution
        y[i] = (x[i] - mu) / sigma
        if y[i] <= -1 * beta:
            y[i] = n * const_a * ((const_b - y[i]) ** (1 - m)) / (m - 1)

        # If it isn't in the power law, then it is Gaussian
        else:
            y[i] = const_a * n * ((const_b + beta) ** (1 - m)) / (m - 1) + n * np.sqrt(
                np.pi / 2
            ) * (erf(beta / np.sqrt(2)) + erf(y[i] / np.sqrt(2)))
    return y


@nb.njit(**nb_defaults(parallel=False))
def nb_crystal_ball_scaled_pdf(
    x: np.ndarray, area: float, mu: float, sigma: float, beta: float, m: float
) -> np.ndarray:
    r"""
    Scaled PDF of a power-law tail plus gaussian.
    As a Numba vectorized function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    area
        The number of counts in the distribution
    mu
        The amount to shift the distribution
    sigma
        The amount to scale the distribution
    beta
        The point where the pdf changes from power-law to Gaussian
    m
        The power of the power-law tail
    """

    return area * nb_crystal_ball_pdf(x, mu, sigma, beta, m)


@nb.njit(**nb_defaults(parallel=False))
def nb_crystal_ball_scaled_cdf(
    x: np.ndarray, area: float, mu: float, sigma: float, beta: float, m: float
) -> np.ndarray:
    r"""
    Scaled CDF for power-law tail plus gaussian. Used for extended binned fits.
    As a Numba vectorized function, it runs slightly faster than
    'out of the box' functions.

    Parameters
    ----------
    x
        The input data
    area
        The number of counts in the distribution
    mu
        The amount to shift the distribution
    sigma
        The amount to scale the distribution
    beta
        The point where the cdf changes from power-law to Gaussian
    m
        The power of the power-law tail
    """

    return area * nb_crystal_ball_cdf(x, mu, sigma, beta, m)


class CrystalBallGen(PygamaContinuous):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x_lo = -1 * np.inf
        self.x_hi = np.inf

    def _pdf(
        self, x: np.ndarray, mu: float, sigma: float, beta: float, m: float
    ) -> np.ndarray:
        x.flags.writeable = True
        return nb_crystal_ball_pdf(x, mu[0], sigma[0], beta[0], m[0])

    def _cdf(
        self, x: np.ndarray, mu: float, sigma: float, beta: float, m: float
    ) -> np.ndarray:
        x.flags.writeable = True
        return nb_crystal_ball_cdf(x, mu[0], sigma[0], beta[0], m[0])

    def get_pdf(
        self, x: np.ndarray, mu: float, sigma: float, beta: float, m: float
    ) -> np.ndarray:
        return nb_crystal_ball_pdf(x, mu, sigma, beta, m)

    def get_cdf(
        self, x: np.ndarray, mu: float, sigma: float, beta: float, m: float
    ) -> np.ndarray:
        return nb_crystal_ball_cdf(x, mu, sigma, beta, m)

    def pdf_norm(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        beta: float,
        m: float,
    ) -> np.ndarray:
        return self._pdf_norm(x, x_lo, x_hi, mu, sigma, beta, m)

    def cdf_norm(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        mu: float,
        sigma: float,
        beta: float,
        m: float,
    ) -> np.ndarray:
        return self._cdf_norm(x, x_lo, x_hi, mu, sigma, beta, m)

    def pdf_ext(
        self,
        x: np.ndarray,
        x_lo: float,
        x_hi: float,
        area: float,
        mu: float,
        sigma: float,
        beta: float,
        m: float,
    ) -> np.ndarray:
        return np.diff(
            nb_crystal_ball_scaled_cdf(np.array([x_lo, x_hi]), area, mu, sigma, beta, m)
        )[0], nb_crystal_ball_scaled_pdf(x, area, mu, sigma, beta, m)

    def cdf_ext(
        self, x: np.ndarray, area: float, mu: float, sigma: float, beta: float, m: float
    ) -> np.ndarray:
        return nb_crystal_ball_scaled_cdf(x, area, mu, sigma, beta, m)

    def required_args(self) -> tuple[str, str, str, str]:
        return "mu", "sigma", "beta", "m"


crystal_ball = CrystalBallGen(name="crystal_ball")



import numpy as np
from scipy.stats import uniform as scipy_uniform

from pygama.math.functions.linear import linear
from pygama.math.functions.pygama_continuous import PygamaContinuous
from pygama.math.functions.uniform import uniform


def test_linear_pdf():
    x = np.arange(-10, 12)
    m = 0
    b = 2
    x_lo = np.amin(x)
    x_hi = np.amax(x)

    y = linear.pdf(x, x_lo, x_hi, m, b)
    y_direct = linear.get_pdf(x, x_lo, x_hi, m, b)
    scipy_y = scipy_uniform.pdf(x, x[0], x[-1] + np.abs(x[0]))

    assert isinstance(linear, PygamaContinuous)
    assert np.array_equal(y, scipy_y)
    assert np.array_equal(y_direct, scipy_y)

    n_sig = 20
    x_lo = -10.0
    x_hi = 11.0

    y_sig, y_ext = linear.pdf_ext(x, x_lo, x_hi, n_sig, m, b)
    assert np.array_equal(y_ext, n_sig * scipy_y)
    assert np.array_equal(y_sig, n_sig)

    norm_y = linear.pdf_norm(x, x[0], x[-1], m, b)
    assert np.array_equal(norm_y, scipy_y)


def test_linear_cdf():
    x = np.arange(-10, 12)
    m = 0
    b = 2
    x_lo = np.amin(x)
    x_hi = np.amax(x)

    y = linear.cdf(x, x_lo, x_hi, m, b)
    y_direct = linear.get_cdf(x, x_lo, x_hi, m, b)
    scipy_y = scipy_uniform.cdf(x, x[0], x[-1] + np.abs(x[0]))

    assert isinstance(uniform, PygamaContinuous)
    assert np.allclose(y, scipy_y, rtol=1e-8)
    assert np.allclose(y_direct, scipy_y, rtol=1e-8)

    n_sig = 20
    y_ext = linear.cdf_ext(x, x_lo, x_hi, n_sig, m, b)
    assert np.allclose(y_ext, n_sig * scipy_y, rtol=1e-8)

    norm_y = linear.cdf_norm(x, x[0], x[-1], m, b)
    assert np.allclose(norm_y, scipy_y)


def test_required_args():
    names = linear.required_args()
    assert names[0] == "x_lo"
    assert names[1] == "x_hi"
    assert names[2] == "m"
    assert names[3] == "b"


def test_name():
    assert linear.name == "linear"
